"""
Note SASP/SBSP has data stored in two files with a range of years
The mocked data is provided to test but is cropped to shorten the files

Double Note: DOY 1 == Jan 1

Mocked data has been cropped to:

* SASP/SBSP 2003-2009 -> 2009 DOY 60-80
* SASP/SBSP 2010-2023 -> 2023 DOY 60-80
* SBSG - 2023 DOY 100-120
* PTSP - 2023 DOY 60-80

"""
import pytest

from metloom.pointdata import CSASMet
from metloom.pointdata.csas import InvalidDateRange
from metloom.variables import CSASVariables
from datetime import datetime, timedelta
from pathlib import Path
import shutil
from unittest.mock import patch, MagicMock
import requests


DATA_DIR = str(Path(__file__).parent.joinpath("data/csas_mocks"))

# Convenient Dates for testing
DT_20090301 = datetime(2009, 3, 1)
DT_20090315 = datetime(2009, 3, 15)
DT_20230301 = datetime(2023, 3, 1)
DT_20230315 = datetime(2023, 3, 15)
DT_20230601 = datetime(2023, 6, 1)
DT_20230615 = datetime(2023, 6, 15)


class TestCSASMet:

    @classmethod
    def get_side_effect(cls, *args, **kwargs):
        url = Path(args[0])
        local_path = Path(__file__).parent.joinpath(DATA_DIR).joinpath(url.name)

        obj = MagicMock()
        with open(local_path, 'rb') as f:
            lines = f.readlines()

        obj.iter_lines.return_value = lines
        obj.__enter__.return_value = obj  # Black magic?
        obj.__exit__.return_value = None
        return obj

    @pytest.fixture(scope="class")
    def mocked_requests(self):
        with patch("metloom.pointdata.files.requests.get") as mock_get:
            mock_get.side_effect = self.get_side_effect
            yield mock_get

    def copy_files(self, urls):
        files = []
        for url in urls:
            file = Path(DATA_DIR).joinpath(Path(url).name)
            cache = Path(__file__).parent.joinpath('cache')
            shutil.copy(file, cache.joinpath(file.name))
            files.append(file)
        return files

    @pytest.fixture(scope='function')
    def cache_dir(self):
        """Cachae dir where data is being downloaded to"""
        cache = Path(__file__).parent.joinpath('cache')
        yield cache
        if cache.is_dir():
            shutil.rmtree(cache)

    @pytest.fixture(scope='function')
    def station(self, mocked_requests, cache_dir, station_id):
        pnt = CSASMet(station_id, cache=cache_dir)
        yield pnt

    @pytest.mark.parametrize('year, doy, hour, expected', [
        (2024, 92, 1400, datetime(2024, 4, 1, 14)),
        # Check doy 1 is jan 1
        (2024, 1, 0, datetime(2024, 1, 1))

    ])
    def test_parse_datetime(self, year, doy, hour, expected):
        dt = CSASMet._parse_datetime({'Year': year, 'DOY': doy, 'Hour': hour})
        assert dt == expected

    @pytest.mark.parametrize("station_id, start, end, expected", [
        # Two stations have different files depending on the years
        ('SBSP', datetime(2003, 1, 1),
         datetime(2003, 1, 1), ['SBSP_1hr_2003-2009.csv']),
        ('SBSP', datetime(2010, 1, 1),
         datetime(2010, 1, 1), ['SBSP_1hr_2010-2024.csv']),
        # Test straddling the files
        ('SBSP', datetime(2008, 1, 1),
         datetime(2010, 1, 1), ['SBSP_1hr_2003-2009.csv', 'SBSP_1hr_2010-2024.csv']),
        # test stations that don't have year specific timeframes
        ('PTSP', datetime(2010, 1, 1), datetime(2010, 1, 2), ["PTSP_1hr.csv"]),
        ('SBSG', datetime(2010, 1, 1), datetime(2010, 1, 2), ["SBSG_1hr.csv"])
    ])
    def test_file_urls(self, station_id, start, end, expected):
        pnt = CSASMet(station_id)
        pnt._verify_station()
        urls = pnt._file_urls(station_id, start, end)
        names = sorted([Path(url).name for url in urls])
        assert names == expected

    @pytest.mark.parametrize("station_id, start, end", [
        # Test pre 2003
        ('SBSP', datetime(2002, 1, 1),
         datetime(2003, 1, 1)),
        # Test post 2023 exception
        ('SBSP', datetime(2010, 1, 1),
         datetime(2030, 1, 1))
    ])
    def test_file_urls_exception(self, station_id, start, end):
        """
        With files there are hard timeframes, test ane exception is raised when
        this is the case
        """
        pnt = CSASMet(station_id)
        pnt._verify_station()
        with pytest.raises(InvalidDateRange):
            pnt._file_urls(station_id, start, end)

    @pytest.mark.parametrize("station_id, year", [
        # Test the two SBSP urls
        ('SBSP', 2009),
        ('SBSP', 2010),
        # Test the two SASP urls
        ('SASP', 2009),
        ('SASP', 2010),
        # Test the Stream gauges url
        ('SBSG', 2005),
        # Test the PTSP url
        ('PTSP', 2003),
    ])
    def test_links_are_valid(self, station_id, year):
        """
        Seeking answers from CSAS on how and when these files are updated. Until then
        this will serve as a nice way to check the files are workin still
        """
        start = datetime(year, 1, 1)
        end = start + timedelta(days=1)

        pnt = CSASMet(station_id)
        pnt._verify_station()
        urls = pnt._file_urls(station_id, start, end)

        resp = requests.head(urls[0])
        assert resp.ok

    @pytest.mark.parametrize('station_id, variable, start, end, expected_mean', [
        ('SASP', CSASVariables.SURF_TEMP, DT_20230301, DT_20230315, -11.59220),
        # Span two files, Test data wont be complete but requires two files
        ('SASP', CSASVariables.SURF_TEMP, DT_20090301, DT_20230315, -10.35284),
    ])
    def test_get_daily_data(self, station, station_id, variable, start, end,
                            expected_mean):
        """ Check pulling two weeks of data """

        df = station.get_daily_data(start, end, [variable])
        # Assert it's a daily timeseries
        assert df.index.get_level_values('datetime').inferred_freq == 'D'
        assert df[variable.name].mean() == pytest.approx(expected_mean, abs=1e-5)

    @pytest.mark.parametrize('station_id, variable, start, end, expected_mean', [
        ('SASP', CSASVariables.SURF_TEMP, DT_20230301, DT_20230315, -11.49969),
    ])
    def test_get_hourly_data(self, station, station_id, variable, start, end,
                             expected_mean):
        """ Check pulling two weeks of data """

        df = station.get_hourly_data(start, end, [variable])

        # Assert it's a daily timeseries
        assert df.index.get_level_values('datetime').inferred_freq.lower() == 'h'
        assert df[variable.name].mean() == pytest.approx(expected_mean, abs=1e-5)



import json
from datetime import datetime
from pathlib import Path
import geopandas as gpd
import pandas as pd
import pytest
from unittest.mock import patch, MagicMock

from metloom.pointdata import norway
from metloom.variables import MetNorwayVariables


class TestMetNorway:
    MOCKS_DIR = Path(__file__).parent.joinpath("data/frost_mocks")

    @classmethod
    def _get_side_effect(cls, *args, **kwargs):
        """
        Side effect for requests.get
        """
        mock_resp = MagicMock()
        # mock observations endpoint
        if "observations/v0" in args[0]:

            if "air_temperature" in kwargs["params"]["elements"]:
                # Success case
                mock_resp.status_code = 200
                with open(cls.MOCKS_DIR.joinpath("hourly_temp.json")) as fp:
                    obj = json.load(fp)
                mock_resp.json.return_value = obj
            else:
                # Case of no data
                mock_resp.status_code = 412

        # mock metadata endpoint
        elif "sources/v0" in args[0]:
            params = kwargs["params"]
            mock_resp.status_code = 200
            with open(cls.MOCKS_DIR.joinpath("search.json")) as fp:
                obj = json.load(fp)
            ids = params.get("ids")
            if ids:
                # filter to ids
                data = obj["data"]
                data = [d for d in data if d["id"] in ids]
                obj["data"] = data

            mock_resp.json.return_value = obj
        else:
            raise NotImplementedError("No other method implemented")
        return mock_resp

    @staticmethod
    def _post_side_effect(*args, **kwargs):
        url = args[0]
        if "auth/accessToken" in url:
            obj = MagicMock()
            obj.json.return_value = {
                "access_token": "FAKE",
                "expires_in": 3600
            }
        else:
            raise NotImplementedError("No other method implemented")
        return obj

    @pytest.fixture(scope='session')
    def token_file(self):
        """
        Json token file fixture for mocking having a token
        """
        d = {
            'client_id': '####',
            'client_secret': '####'
        }
        json_file = Path(__file__).parent.joinpath('frost_token.json')

        with open(json_file, 'w+') as fp:
            json.dump(d, fp)

        yield json_file
        # Clean up
        if json_file.is_file():
            json_file.unlink()

    @pytest.fixture(scope="class")
    def mock_request(self):
        with patch("metloom.pointdata.norway.requests") as mr:
            mr.get.side_effect = self._get_side_effect
            mr.post.side_effect = self._post_side_effect
            yield mr

    @pytest.fixture(scope="class")
    def obj(self, mock_request, token_file):
        yield norway.MetNorwayPointData(
            "SN46432", "BASURA", token_json=token_file
        )

    def test_daily_data(self, obj):
        result = obj.get_daily_data(
            datetime(2023, 8, 1), datetime(2023, 8, 5),
            [MetNorwayVariables.TEMP]
        )
        assert result["AIR TEMP"].values == pytest.approx(
            [13.44583333, 14.325, 13.86666667, 11.64166667]
        )
        assert all(result["AIR TEMP_units"] == ["degC"] * 4)
        assert all(result["quality_code"] == "resampled")

    def test_hourly_data(self, obj):
        result = obj.get_hourly_data(
            datetime(2023, 8, 1), datetime(2023, 8, 5),
            [MetNorwayVariables.TEMP]
        )

        assert len(result) == 96
        assert all(result["AIR TEMP_units"] == ["degC"] * 96)
        assert result["AIR TEMP"].mean() == pytest.approx(13.319792)

    def test_daily_nodata(self, obj):
        result = obj.get_daily_data(
            datetime(2023, 8, 1), datetime(2023, 8, 5),
            [MetNorwayVariables.SWE]
        )
        assert result is None

    def test_points_from_geometry(self, mock_request, token_file):
        shp = gpd.read_file(
            self.MOCKS_DIR.joinpath("box.shp")
        )
        result = norway.MetNorwayPointData.points_from_geometry(
            shp, [MetNorwayVariables.TEMP], token_json=token_file
        )
        assert len(result.points) == 16
        assert result.points[0].id == "SN46432"

    @pytest.mark.parametrize(
        "ref_time, offset, resolution, timeseries_id, expected", [
            ("2023-08-01T00:00:00.000Z", "PT0H", "PT1H", 0,
             "2023-08-01T00:00:00.000Z"),
            ("2023-08-01T00:00:00.000Z", "PT6H", "PT1H", 0,
             "2023-08-01T06:00:00.000Z"),
            ("2023-08-01T00:00:00.000Z", "PT6H", "PT1H", 1,
             "2023-08-01T07:00:00.000Z"),
            ("2023-08-01T00:00:00.000Z", "PT6H", "PT12H", 1,
             "2023-08-01T18:00:00.000Z"),
        ]
    )
    def test_observation_time(
        self, ref_time, offset, resolution, timeseries_id, expected
    ):
        result = norway.MetNorwayPointData._time_info_to_observation_time(
            ref_time, offset, resolution, timeseries_id
        )
        assert result == pd.to_datetime(expected)



"""
A reader for the Mammoth CUES site
https://snow.ucsb.edu/index.php/description/

"""
from datetime import datetime, timezone, timedelta
from io import StringIO
from typing import List
import logging
import geopandas as gpd
import pandas as pd
import requests

from metloom.pointdata import PointData
from metloom.variables import CuesLevel1Variables, SensorDescription

LOG = logging.getLogger(__name__)


class CuesLevel1(PointData):
    """
    Implement PointData methods for CUES level 1 data
    https://snow.ucsb.edu/index.php/description/
    https://snow.ucsb.edu/index.php/query-db/
    https://doi.org/10.21424/R4159Q

    """
    ALLOWED_VARIABLES = CuesLevel1Variables
    URL = "https://snow.ucsb.edu/index.php/query-db/"
    DATASOURCE = "UCSB CUES"

    def __init__(self, station_id, name, metadata=None):
        """
        See docstring for PointData.__init__
        """
        super(CuesLevel1, self).__init__(
            station_id or "CUES",
            name or "CUES",
            metadata=metadata
        )
        self._raw_metadata = None
        self._tzinfo = timezone(timedelta(hours=-8.0))

    def _get_one_vaiable(
        self, start_date, end_date, variables: SensorDescription,
        period, method
    ):
        dt_fmt = "%Y-%m-%d"
        data = dict(
            # table="downward looking solar radiation",
            table=variables.code, start=start_date.strftime(dt_fmt),
            end=end_date.strftime(dt_fmt), interval=period,
            method=method, output="CSV",
            category="Measurement"
        )
        resp = requests.post(self.URL, data=data)
        resp.raise_for_status()
        return resp.content.decode()

    def _sensor_response_to_df(self, data, variable):

        # Check for no data
        if not data.replace("\n", ""):
            LOG.debug(f"No data returned for {variable}")
            return None

        # Parse the 'csv' string returned
        df = pd.read_csv(
            StringIO(data), delimiter=",", skip_blank_lines=True,
            comment="#"
        )
        columns = list(df.columns.values)
        # check that we have the expected columns for the
        # instrument since multiple may be returned
        if variable.instrument:
            var_column = None
            for c in columns:
                if variable.instrument in c:
                    var_column = c
                    break
            if var_column is None:
                LOG.error(f"Returned columns were {columns}")
                raise RuntimeError(
                    f"Could not find column for expected"
                    f" instrument {variable.instrument}"
                )
        elif len(columns) > 2:
            raise RuntimeError(
                f"Expected 2 columns, got {columns}"
            )
        else:
            # just use the second of two columns
            var_column = columns[1]
        column_map = {
            columns[0]: "datetime",
            var_column: variable.name
        }
        # Parse the units out of the returned column name
        units = columns[1].split(";")[-1].replace(
            "(", ""
        ).replace(")", "").strip()
        # Rename to desired columns and add a units column
        df.rename(columns=column_map, inplace=True)

        # handle the timezone convert to UTC
        df["datetime"] = pd.to_datetime(df["datetime"])
        df["datetime"] = df["datetime"].apply(self._handle_df_tz)

        df = df.set_index("datetime")
        df = df.loc[:, [variable.name]]
        df[f"{variable.name}_units"] = [units] * len(df)

        return df

    def _get_data(
        self, start_date, end_date, variables: List[SensorDescription],
        period,
    ):
        df = pd.DataFrame()
        df.index.name = "datetime"
        for variable in variables:
            method = "sum" if variable.accumulated else "average"
            data = self._get_one_vaiable(
                start_date, end_date, variable, period, method
            )
            df_var = self._sensor_response_to_df(data, variable)
            if df_var is not None:
                df[df_var.columns] = df_var
        # Set the site info
        df["site"] = [self.id] * len(df)
        df["datasource"] = [self.DATASOURCE] * len(df)
        # Make this a geodataframe
        df = gpd.GeoDataFrame(df, geometry=[self.metadata] * len(df))
        df = df.reset_index().set_index(["datetime", "site"])
        self.validate_sensor_df(df)
        return df

    def get_daily_data(self, start_date: datetime, end_date: datetime,
                       variables: List[SensorDescription]):
        return self._get_data(
            start_date, end_date, variables, "day"
        )

    def get_hourly_data(self, start_date: datetime, end_date: datetime,
                        variables: List[SensorDescription]):
        return self._get_data(
            start_date, end_date, variables, "hr"
        )

    def get_snow_course_data(self, start_date: datetime, end_date: datetime,
                             variables: List[SensorDescription]):
        raise NotImplementedError("Not implemented")

    def _get_metadata(self):
        pass

    def points_from_geometry(self, geometry: gpd.GeoDataFrame,
                             variables: List[SensorDescription],
                             snow_courses=False, within_geometry=True,
                             buffer=0.0):
        raise NotImplementedError("Not implemented")

    @property
    def metadata(self):
        """
        Hardcode the metadata
        """
        return gpd.points_from_xy(
            [-119.029128], [37.643093], [9661]
        )[0]



import json
from os.path import join
from pathlib import Path
from unittest.mock import patch, MagicMock

import geopandas as gpd
import pandas as pd
import pytest
from shapely.geometry import Point

from metloom.pointdata import NWSForecastPointData
from metloom.variables import NWSForecastVariables
from tests.test_point_data import BasePointDataTest

DATA_DIR = str(Path(__file__).parent.joinpath("data/nws_mocks"))


class TestNWSForecast(BasePointDataTest):
    @classmethod
    def get_side_effect(cls, *args, **kwargs):
        url = args[0]
        if ".gov/gridpoints" in url:
            with open(join(DATA_DIR, "meta_and_data.json")) as fp:
                data = json.load(fp)
        elif ".gov/points" in url:
            with open(join(DATA_DIR, "initial_meta.json")) as fp:
                data = json.load(fp)
        else:
            raise RuntimeError(f"{url} is an unknown option")

        obj = MagicMock()
        obj.json.return_value = data
        return obj

    @pytest.fixture(scope="class")
    def mocked_requests(self):
        with patch("requests.get") as mock_get:
            mock_get.side_effect = self.get_side_effect
            yield mock_get

    @pytest.fixture(scope="function")
    def station(self, mocked_requests):
        point1 = Point(-119, 43)
        pt = NWSForecastPointData(
            "test", None, initial_metadata=point1
        )
        yield pt

    @pytest.fixture(scope="class")
    def expected_meta(self):
        return Point(-118.99345926915265, 42.99291053264557, 5000.00016)

    @pytest.fixture(scope="class")
    def daily_expected(self, expected_meta):
        dts = [
            '2024-06-25T00:00:00+0000', '2024-06-26T00:00:00+0000',
            '2024-06-27T00:00:00+0000', '2024-06-28T00:00:00+0000',
            '2024-06-29T00:00:00+0000', '2024-06-30T00:00:00+0000',
            '2024-07-01T00:00:00+0000', '2024-07-02T00:00:00+0000',
            '2024-07-03T00:00:00+0000'
        ]
        temp_values = [
            26.82539683, 23.647343, 16.18357488, 14.19191919, 19.46859903,
            18.68686869, 16.06280193, 16.61835749, 27.22222222]
        df = pd.DataFrame()
        df["datetime"] = pd.to_datetime(dts)
        df["AIR TEMP"] = temp_values
        df['AIR TEMP_units'] = ["degC"] * len(df)
        df["site"] = ["test"] * len(df)
        df["datasource"] = ["NWS Forecast"] * len(df)
        df = gpd.GeoDataFrame(df, geometry=[expected_meta] * len(df))
        # needed to reorder the columns for the pd testing compare
        df = df.filter(
            [
                "datetime",
                "geometry",
                "AIR TEMP",
                "site",
                "AIR TEMP_units",
                "datasource",
            ]
        )
        df.set_index(keys=["datetime", "site"], inplace=True)
        return df

    def test_get_metadata(self, station, expected_meta):
        result = station.metadata
        assert expected_meta == result

    def test_get_daily_data(self, station, daily_expected):
        response = station.get_daily_forecast(
            [NWSForecastVariables.TEMP],
        )
        pd.testing.assert_frame_equal(
            response.sort_index(axis=1),
            daily_expected.sort_index(axis=1)
        )



from datetime import timezone, timedelta, datetime
from unittest.mock import MagicMock, patch
import re

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import shapely
from pandas import Timestamp
from pathlib import Path

from metloom.pointdata import CDECPointData, PointDataCollection
from metloom.variables import CdecStationVariables
from tests.test_point_data import BasePointDataTest, side_effect_error


class TestCDECStation(BasePointDataTest):
    CDEC_MOCKS_DIR = Path(__file__).parent.joinpath("data/cdec_mocks")

    @staticmethod
    def cdec_daily_precip_response():
        return [
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 2,
                "sensorType": "SNOW WC",
                "date": "2021-5-16 00:00",
                "obsDate": "2021-5-16 00:00",
                "value": -0.11,
                "dataFlag": " ",
                "units": "INCHES",
            },
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 2,
                "sensorType": "SNOW WC",
                "date": "2021-5-17 00:00",
                "obsDate": "2021-5-17 00:00",
                "value": -0.10,
                "dataFlag": " ",
                "units": "INCHES",
            },
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 2,
                "sensorType": "SNOW WC",
                "date": "2021-5-18 00:00",
                "obsDate": "2021-5-18 00:00",
                "value": -0.10,
                "dataFlag": " ",
                "units": "INCHES",
            },
        ]

    @staticmethod
    def cdec_daily_temp_response():
        return [
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 30,
                "sensorType": "SNOW WC",
                "date": "2021-5-15 00:00",
                "obsDate": "2021-5-15 00:00",
                "value": 2.1,
                "dataFlag": " ",
                "units": "DEG F",
            },
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 30,
                "sensorType": "SNOW WC",
                "date": "2021-5-17 00:00",
                "obsDate": "2021-5-17 00:00",
                "value": 2.4,
                "dataFlag": " ",
                "units": "DEG F",
            },
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 30,
                "sensorType": "SNOW WC",
                "date": "2021-5-18 00:00",
                "obsDate": "2021-5-18 00:00",
                "value": 2.2,
                "dataFlag": " ",
                "units": "DEG F",
            },
        ]

    @staticmethod
    def cdec_hourly_temp_response():
        return [
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 30,
                "sensorType": "SNOW WC",
                "date": "2021-5-15 00:00",
                "obsDate": "2021-5-15 00:00",
                "value": 2.1,
                "dataFlag": " ",
                "units": "DEG F",
            },
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 30,
                "sensorType": "SNOW WC",
                "date": "2021-5-15 01:00",
                "obsDate": "2021-5-15 01:00",
                "value": 2.4,
                "dataFlag": " ",
                "units": "DEG F",
            },
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 30,
                "sensorType": "SNOW WC",
                "date": "2021-5-15 03:00",
                "obsDate": "2021-5-15 03:00",
                "value": 2.2,
                "dataFlag": " ",
                "units": "DEG F",
            },
        ]

    @pytest.fixture(scope="function")
    def tny_station(self, mock_get):
        yield CDECPointData("TNY", "Tenaya Lake")

    @pytest.fixture(scope="function")
    def tny_station_force_hourly(self, mock_get_force_hourly):
        yield CDECPointData("TNY", "Tenaya Lake")

    @pytest.fixture(scope="class")
    def expected_points(self):
        return gpd.points_from_xy([-119.449875], [37.837581], z=[8150.0])

    @pytest.fixture(scope="class")
    def tny_daily_expected(self, expected_points):
        df = gpd.GeoDataFrame.from_dict(
            [
                {
                    "datetime": pd.Timestamp("2021-05-15 08:00:00+0000",
                                             tz="UTC"),
                    "ACCUMULATED PRECIPITATION": np.nan,
                    "ACCUMULATED PRECIPITATION_units": np.nan,
                    "AVG AIR TEMP": 2.1,
                    "AVG AIR TEMP_units": "DEG F",
                    "site": "TNY",
                    "datasource": "CDEC"
                },
                {
                    "datetime": pd.Timestamp("2021-05-16 08:00:00+0000", tz="UTC"),
                    "ACCUMULATED PRECIPITATION": -0.11,
                    "ACCUMULATED PRECIPITATION_units": "INCHES",
                    "AVG AIR TEMP": np.nan,
                    "AVG AIR TEMP_units": np.nan,
                    "site": "TNY",
                    "datasource": "CDEC"
                },
                {
                    "datetime": pd.Timestamp("2021-05-17 08:00:00+0000", tz="UTC"),
                    "ACCUMULATED PRECIPITATION": -0.10,
                    "ACCUMULATED PRECIPITATION_units": "INCHES",
                    "AVG AIR TEMP": 2.4,
                    "AVG AIR TEMP_units": "DEG F",
                    "site": "TNY",
                    "datasource": "CDEC"
                },
                {
                    "datetime": pd.Timestamp("2021-05-18 08:00:00+0000", tz="UTC"),
                    "ACCUMULATED PRECIPITATION": -0.10,
                    "ACCUMULATED PRECIPITATION_units": "INCHES",
                    "AVG AIR TEMP": 2.2,
                    "AVG AIR TEMP_units": "DEG F",
                    "site": "TNY",
                    "datasource": "CDEC"
                },
            ],
            geometry=[expected_points[0]] * 4,
        )
        # needed to reorder the columns for the pd testing compare
        df = df.filter(
            [
                "datetime",
                "geometry",
                "site",
                "measurementDate",
                "ACCUMULATED PRECIPITATION",
                "ACCUMULATED PRECIPITATION_units",
                "AVG AIR TEMP",
                "AVG AIR TEMP_units",
                "datasource"
            ]
        )
        df.set_index(keys=["datetime", "site"], inplace=True)
        return df

    @classmethod
    def tny_meta_return(cls):
        with open(cls.CDEC_MOCKS_DIR.joinpath("raw_tny_locations.csv"), "r") as fp:
            lines = fp.readlines()
        return "".join(lines)

    def read_html_side_effect(self, url, **kwargs):
        if "dynamicapp/staSearch" in url:
            return self.station_search_side_effect(url)
        else:
            raise ValueError("Unknown html scenario in the test mocks")

    @pytest.fixture
    def mock_read_html(self):
        with patch("metloom.pointdata.cdec.pd.read_html") as mock_table_read:
            mock_table_read.side_effect = self.read_html_side_effect
            yield mock_table_read

    @classmethod
    def get_side_effect(cls, url, **kwargs):
        mock = MagicMock()
        params = kwargs.get("params")
        if params:
            if params.get("dur_code") == "D" and params.get('SensorNums') == "2":
                mock.json.return_value = cls.cdec_daily_precip_response()
            elif params.get("dur_code") == "D" and params.get('SensorNums') == "30":
                mock.json.return_value = cls.cdec_daily_temp_response()
            elif params.get("dur_code") == 'h':
                mock.json.return_value = cls.cdec_hourly_temp_response()
            else:
                raise NotImplementedError("Not implemented")
        # return the metadata
        elif "CSVMetaDataServlet" in url:
            mock.text = cls.tny_meta_return()
        else:
            mock.json.return_value = []
        return mock

    @classmethod
    def get_side_effect_just_hourly(cls, url, **kwargs):
        """
        Mock to force an hourly return for testing resample
        """
        mock = MagicMock()
        params = kwargs.get("params")
        if params:
            if params.get("dur_code") == 'D':
                mock.json.return_value = []
            elif params.get("dur_code") == 'h':
                mock.json.return_value = cls.cdec_hourly_temp_response()
            else:
                raise NotImplementedError("Not implemented")
        # return the metadata
        elif "CSVMetaDataServlet" in url:
            mock.text = cls.tny_meta_return()
        else:
            mock.json.return_value = []
        return mock

    @pytest.fixture()
    def mock_get(self):
        with patch("metloom.pointdata.cdec.requests") as mock_requests:
            obj = mock_requests.get
            obj.side_effect = self.get_side_effect
            yield obj

    @pytest.fixture()
    def mock_get_force_hourly(self):
        """
        Mock to force an hourly return for testing resample
        """
        with patch("metloom.pointdata.cdec.requests") as mock_requests:
            obj = mock_requests.get
            obj.side_effect = self.get_side_effect_just_hourly
            yield obj

    @classmethod
    def station_search_side_effect(cls, *args, **kargs):
        url = args[0]
        sensor_num = re.findall(r'.*&sensor=(\d+)&', url)[0]
        if sensor_num == "3":
            return cls.station_search_response()
        elif sensor_num == "18":
            return [
                pd.DataFrame.from_records(
                    [
                        (
                            "AAA",
                            "A Fake Station",
                            "TUOLUMNE R",
                            "TUOLUMNE",
                            -119.0,
                            37.0,
                            9900,
                            "CA Dept of Water Resources/DFM-Hydro-SMN",
                            np.nan,
                        ),
                        (
                            "DAN",
                            "DANA MEADOWS",
                            "TUOLUMNE R",
                            "TUOLUMNE",
                            -119.257,
                            37.897,
                            9800,
                            "CA Dept of Water Resources/DFM-Hydro-SMN",
                            np.nan,
                        ),
                        (
                            "TNY",
                            "TENAYA LAKE",
                            "MERCED R",
                            "MARIPOSA",
                            -119.448,
                            37.838,
                            8150,
                            "CA Dept of Water Resources/DFM-Hydro-SMN",
                            np.nan,
                        ),
                        (
                            "BBB",
                            "B Fake Station",
                            "TUOLUMNE R",
                            "TUOLUMNE",
                            -119.5,
                            37.5,
                            9905,
                            "CA Dept of Water Resources/DFM-Hydro-SMN",
                            np.nan,
                        ),
                    ],
                    columns=[
                        "ID",
                        "Station Name",
                        "River Basin",
                        "County",
                        "Longitude",
                        "Latitude",
                        "ElevationFeet",
                        "Operator",
                        "Map",
                    ],
                )
            ]
        else:
            raise ValueError(f"{sensor_num} is not configured")

    @staticmethod
    def station_search_response():
        return [
            pd.DataFrame.from_records(
                [
                    (
                        "GIN",
                        "GIN FLAT",
                        "MERCED R",
                        "MARIPOSA",
                        -119.773,
                        37.767,
                        7050,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                    (
                        "DAN",
                        "DANA MEADOWS",
                        "TUOLUMNE R",
                        "TUOLUMNE",
                        -119.257,
                        37.897,
                        9800,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                    (
                        "TNY",
                        "TENAYA LAKE",
                        "MERCED R",
                        "MARIPOSA",
                        -119.448,
                        37.838,
                        8150,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                    (
                        "GFL",
                        "GIN FLAT (COURSE)",
                        "MERCED R",
                        "MARIPOSA",
                        -119.773,
                        37.765,
                        7000,
                        "Yosemite National Park",
                        np.nan,
                    ),
                    (
                        "TUM",
                        "TUOLUMNE MEADOWS",
                        "TUOLUMNE R",
                        "TUOLUMNE",
                        -119.350,
                        37.873,
                        8600,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                    (
                        "SLI",
                        "SLIDE CANYON",
                        "TUOLUMNE R",
                        "TUOLUMNE",
                        -119.43,
                        38.092,
                        9200,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                ],
                columns=[
                    "ID",
                    "Station Name",
                    "River Basin",
                    "County",
                    "Longitude",
                    "Latitude",
                    "ElevationFeet",
                    "Operator",
                    "Map",
                ],
            )
        ]

    def test_class_variables(self):
        assert CDECPointData("no", "no").tzinfo == timezone(timedelta(hours=-8.0))

    def test_get_metadata(self, mock_get, tny_station, expected_points):
        metadata = tny_station.metadata

        mock_get.assert_called_once_with(
            "https://cdec.water.ca.gov/dynamicapp/req/CSVMetaDataServlet?Stations=TNY"
        )

        assert expected_points[0] == metadata

    def test_get_daily_data(
        self, mock_get, mock_read_html, tny_station, tny_daily_expected
    ):
        response = tny_station.get_daily_data(
            datetime(2021, 5, 15),
            datetime(2021, 5, 18),
            [CdecStationVariables.PRECIPITATIONACCUM,
             CdecStationVariables.TEMPAVG],
        )
        # mock_get = mock_requests.get
        mock_get.assert_any_call(
            "http://cdec.water.ca.gov/dynamicapp/req/JSONDataServlet",
            params={
                "Stations": "TNY",
                "dur_code": "D",
                "Start": "2021-05-15T00:00:00",
                "End": "2021-05-18T00:00:00",
                "SensorNums": "30",
            },
        )
        assert mock_get.call_count == 3
        pd.testing.assert_frame_equal(
            response.sort_index(axis=1),
            tny_daily_expected.sort_index(axis=1)
        )

    def test_get_daily_from_hourly_data(
        self, mock_read_html, mock_get_force_hourly, tny_station_force_hourly,
        expected_points
    ):
        """
        Check that we fall back on resampled hourly data if we don't find
        daily data
        """
        response = tny_station_force_hourly.get_daily_data(
            datetime(2021, 5, 15),
            datetime(2021, 5, 16),
            [CdecStationVariables.TEMPAVG],
        )
        assert mock_get_force_hourly.call_count == 3
        expected = gpd.GeoDataFrame.from_dict(
            {
                'AVG AIR TEMP': {
                    (Timestamp('2021-05-15 08:00:00+0000', tz='UTC'),
                        'TNY'): 2.233333},
                'AVG AIR TEMP_units': {
                    (Timestamp('2021-05-15 08:00:00+0000', tz='UTC'),
                        'TNY'): 'DEG F'},
                'datasource': {
                    (Timestamp('2021-05-15 08:00:00+0000', tz='UTC'),
                     'TNY'): 'CDEC'}
            }, geometry=expected_points
        )
        expected.index.set_names(["datetime", "site"], inplace=True)
        pd.testing.assert_frame_equal(
            response, expected, check_exact=False, check_like=True
        )

    def test_points_from_geometry(self, mock_read_html, mock_get, shape_obj):
        expected_url = (
            "https://cdec.water.ca.gov/dynamicapp/staSearch?"
            "sta=&sensor_chk=on&sensor=3"
            "&collect=NONE+SPECIFIED&dur="
            "&active_chk=on&active=Y"
            "&loc_chk=on&lon1=-119.8"
            "&lon2=-119.2&lat1=37.7"
            "&lat2=38.2"
            "&elev1=-5&elev2=99000&nearby=&basin=NONE+SPECIFIED"
            "&hydro=NONE+SPECIFIED&county=NONE+SPECIFIED"
            "&agency_num=160&display=sta"
        )
        result = CDECPointData.points_from_geometry(
            shape_obj, [CdecStationVariables.SWE]
        )
        calls = mock_read_html.call_args_list
        assert calls[0][0][0] == expected_url
        assert len(result) == 6
        assert [st.id for st in result] == [
            "GIN", "DAN", "TNY", "GFL", "TUM", "SLI"
        ]

    @staticmethod
    def check_str_for_float(whole_string, key):
        """
        Parse lat/lon from cdec search string for table read
        """
        split_str = whole_string.split("&")
        matches = [ss for ss in split_str if key in ss]
        if len(matches) > 1 or len(matches) == 0:
            raise ValueError(f"{key} is a bad checker")
        match = matches[0]
        result_val = match.split("=")[-1]
        return float(result_val)

    def test_points_from_geometry_buffer(self, mock_read_html, shape_obj):
        CDECPointData.points_from_geometry(
            shape_obj, [CdecStationVariables.SWE], buffer=0.1
        )

        result_str = mock_read_html.call_args_list[0][0][0]
        expected = {
            'lat2': 38.3, 'lat1': 37.6,
            'lon2': -119.1, 'lon1': -119.9
        }
        for k, v in expected.items():
            result = self.check_str_for_float(result_str, k)
            assert v == pytest.approx(result)

    def test_points_from_geometry_multi_sensor(self, mock_read_html, shape_obj):
        with patch.object(
            CDECPointData, 'is_only_snow_course', return_value=False
        ):
            result = CDECPointData.points_from_geometry(
                shape_obj,
                [CdecStationVariables.SWE, CdecStationVariables.SNOWDEPTH],
                within_geometry=False
            )
            expected_names = [
                "A Fake Station", "B Fake Station", "GIN FLAT",
                "DANA MEADOWS", "TENAYA LAKE", "GIN FLAT (COURSE)",
                "TUOLUMNE MEADOWS", "SLIDE CANYON"
            ]
            expected_codes = [
                "AAA", "BBB", "GIN", "DAN", "TNY", "TUM", "SLI", "GFL"
            ]
            assert len(result) == 8
            assert all([st.id in expected_codes for st in result])
            assert all([st.name in expected_names for st in result])

    def test_points_from_geometry_fail(self, shape_obj):
        with patch("metloom.pointdata.cdec.pd") as mock_pd:
            mock_pd.read_html.side_effect = side_effect_error
            result = CDECPointData.points_from_geometry(
                shape_obj, [CdecStationVariables.SWE]
            )
            assert result.points == []

    def test_point_collection_to_dataframe(self, mock_get, mock_read_html, shape_obj):
        result = CDECPointData.points_from_geometry(
            shape_obj, [CdecStationVariables.SWE]
        )
        assert mock_read_html.call_count == 1
        assert mock_get.call_count == 6
        assert isinstance(result, PointDataCollection)
        points_df = result.to_dataframe()
        for idp, point in enumerate(result):
            point_row = points_df.iloc[idp]
            assert point.name == point_row["name"]
            assert point.id == point_row["id"]
            assert point.metadata == point_row["geometry"]
            assert point.DATASOURCE == point_row["datasource"]

    def test_can_parse_dates(self, tny_station):
        df = pd.DataFrame.from_records([
            {"datetime": "2021-03-14 01:00:00"},
            # This time does not exist in US/Pacific, but does in CDEC
            {"datetime": "2021-03-14 02:00:00"},
            {"datetime": "2021-03-14 03:00:00"},
        ])
        df["datetime"] = pd.to_datetime(df["datetime"])
        df.set_index("datetime", inplace=True)
        df.index = df.index.tz_localize(tny_station.tzinfo)
        df.tz_convert("UTC")


class TestCdecUptime:
    """
    Live tests of cdec
    """
    @pytest.fixture(scope="class")
    def tum(self):
        point = CDECPointData("TUM", "Tuolumne Meadows")
        # call metadata once so we cache it off
        print(point.metadata)
        return point

    def test_metadata(self, tum):
        assert tum.metadata == shapely.geometry.Point(-119.3499675, 37.876215, 8500)

    def test_is_snowcourse(self, tum):
        assert tum.is_partly_snow_course()

    def test_is_only_monthly(self, tum):
        assert not tum.is_only_monthly()

    def test_is_only_snowcourse(self, tum):
        assert not tum.is_only_snow_course([tum.ALLOWED_VARIABLES.SWE])


class TestCdecUptimeTRK:
    @pytest.fixture(scope="class")
    def stn(self):
        point = CDECPointData("TRK", "Truckee River Near Truckee")
        # call metadata once so we cache it off
        print(point.metadata)
        return point

    def test_metadata(self, stn):
        assert stn.metadata == shapely.geometry.Point(
            -120.205474853516, 39.2962951660156, 5858
        )

    def test_is_snowcourse(self, stn):
        assert not stn.is_partly_snow_course()

    def test_is_only_monthly(self, stn):
        assert not stn.is_only_monthly()

    def test_is_only_snowcourse(self, stn):
        assert not stn.is_only_snow_course([stn.ALLOWED_VARIABLES.SWE])



import geopandas as gpd
import argparse

from metloom.dataframe_utils import shp_to_box


def main():
    parser = argparse.ArgumentParser(
        description="Convert shapefile to a box of the shapefile"
    )
    parser.add_argument(
        "shapefile",
        help="shapefile used for conversion"
    )
    args = parser.parse_args()

    df = gpd.read_file(args.shapefile)
    # Calculate the bounding box of the entire GeoDataFrame
    gdf_bbox = shp_to_box(df)

    # Write the bounding box GeoDataFrame to a new shapefile
    gdf_bbox.to_file("austria_box.shp")


if __name__ == '__main__':
    main()



from typing import Union
from datetime import date, datetime
from pathlib import Path

import geopandas as gpd
from shapely.geometry import Point
import pandas as pd
import logging

from .base import PointData, PointDataCollection
from .. import arm_utils
from ..dataframe_utils import shp_to_box
from ..variables import SAILStationVariables, SensorDescription
from ..dataframe_utils import resample_series

LOG = logging.getLogger("metloom.pointdata.sail")


class SAILPointData(PointData):
    """
    https://adc.arm.gov/discovery/#/results/site_code::guc
    """

    ALLOWED_VARIABLES = SAILStationVariables
    DATASOURCE = "SAIL"

    def __init__(
        self,
        station_id: str,
        metadata: dict = None,
        cache: Union[str, Path] = Path(".cache"),
        token_json: Union[str, Path] = Path("~/.arm_token.json"),
    ):
        assert station_id.upper() in ("GUC:M1", "GUC:S1", "GUC:S2", "GUC:S3", "GUC:S4"), (
            f"Invalid station_id: {station_id}"
        )

        super().__init__(
            station_id=station_id.upper(),
            name="Surface Atmosphere Integrated Field Laboratory (SAIL)",
            metadata=metadata,
        )
        self._cache = cache
        self._token_json = Path(token_json).expanduser() if token_json else None

        site = station_id.split(":")
        self._site = site[0].upper()
        self._facility_code = site[1].upper()

    def get_daily_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: list[SensorDescription],
    ):
        self._check_start_end_dates(start_date, end_date)
        return self._download_sail_data(start_date, end_date, variables, interval="D")

    def get_hourly_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: list[SensorDescription],
    ):
        self._check_start_end_dates(start_date, end_date)
        return self._download_sail_data(start_date, end_date, variables, interval="h")

    def _download_sail_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: list[SensorDescription],
        interval: str,
    ) -> pd.DataFrame:
        """
        The ARM data is stored in a series of files based on the sensors at the location.

        This function will download the data for the specified variables and
        return a dataframe with the data. If the files already exist, they will not download
        again.

        NOTE: arm_utils.get_station_data function returns hourly data.
        """
        assert isinstance(variables, list), "variables must be a list of SensorDescription objects"

        columns = []
        for variable in variables:
            sta_site = variable.extra["site"].upper()
            sta_facility_code = variable.extra["facility_code"].upper()
            if not hasattr(self.ALLOWED_VARIABLES, variable.name):
                raise ValueError(f"Variable {variable} is not allowed. Allowed variables are: {self.ALLOWED_VARIABLES}")

            if sta_site != self._site or sta_facility_code != self._facility_code:
                raise ValueError(
                    (
                        f"Variable {variable.code} is not defined for the SAIL site "
                        f"({self._site}:{self._facility_code}), but {sta_site}:{sta_facility_code} provided."
                    )
                )

            df = arm_utils.get_station_data(
                site=variable.extra["site"],
                measurement=variable.extra["measurement"],
                facility_code=variable.extra["facility_code"],
                data_level=variable.extra["data_level"],
                start=start_date,
                end=end_date,
                variables=[variable.code],
                destination=self._cache,
                token_json=self._token_json,
            )
            if df is not None:
                columns.append(pd.Series(resample_series(df[variable.code], variable, interval), name=variable.name))
                units = variable.extra.get("units", None)
                if units is not None:
                    columns.append(pd.Series(units, index=columns[-1].index, name=f"{variable.name}_units"))

        if columns:
            df = pd.concat(columns, axis="columns")
            df["site"] = f"{self._site}:{self._facility_code}"
            df["datasource"] = "ARM"
            df.reset_index(inplace=True)
            df = df.set_index(["datetime", "site"])
            return df
        else:
            LOG.error(
                f"No data found for the specified variables: {', '.join(v.name for v in variables)}.\n"
                f"Please check the variable names and the date range."
            )
        return pd.DataFrame()

    @classmethod
    def points_from_geometry(
        cls,
        geometry: gpd.GeoDataFrame,
        variables: list[SensorDescription],
        snow_courses=None,
        within_geometry=True,
        buffer=0.0,
    ):
        if snow_courses is not None:
            LOG.warning("The snow_courses argument is not used in SAILPointData.points_from_geometry")

        # get geometry object to use for searching within
        boundary = geometry.to_crs(4326) if within_geometry else shp_to_box(geometry).to_crs(4326)
        if buffer > 0:
            boundary = boundary.to_crs(4326).buffer(buffer)

        # get the geometry of the points to check
        stations = list()
        for variable in variables:
            station_id = f"{variable.extra['site']}:{variable.extra['facility_code']}"
            lat, lon, _ = SAILPointData.get_location(station_id, variable)
            stations.append(Point(lon, lat))
        stations = gpd.GeoSeries(stations, crs="EPSG:4326")
        indices = stations[stations.within(boundary)].index.to_list()

        points = [
            SAILPointData(station_id=f"{variables[idx].extra['site']}:{variables[idx].extra['facility_code']}")
            for idx in indices
        ]
        return PointDataCollection(points)

    def get_snow_course_data(
        self,
        start_date: date,
        end_date: date,
        variables: list[SensorDescription],
    ):
        raise NotImplementedError("SAILPointData.get_snow_course_data not implemented")

    def _check_start_end_dates(self, start_date: date, end_date: date):
        """
        Check that the start and end dates are valid
        """
        # get the start and end dates to be date objects for comparison
        start = date.fromisoformat(start_date) if isinstance(start_date, str) else start_date
        end = date.fromisoformat(end_date) if isinstance(end_date, str) else end_date
        start = start.date() if hasattr(start, "date") else start
        end = end.date() if hasattr(end, "date") else end

        # check that the start and end dates are valid
        if start > end:
            raise ValueError("Start date must be before end date")
        if start < date(2021, 9, 1):
            raise ValueError(f"Start date, {start}, must be after 2021-09-01, the first date of data available")
        if end > date(2023, 6, 16):
            raise ValueError(f"End date, {end}, must be before 2023-06-16, the last date of data available")

    @staticmethod
    def get_location(station_id: str, variable: SensorDescription = None) -> tuple[float, float, float]:
        """
        Get the location of the site and facility code.

        The Gunnison SAIL site has 3 supplemental sites (S1, S2, S3, S4) and one main site (M1). The S4 site
        is atmospheric measurements made with a teathered balloon, thus the location is not constant and
        it is excluded from the hard-coded locations.

        Returns a tuple of (latitude, longitude, elevation [m])
        https://www.arm.gov/capabilities/observatories/guc/locations
        """

        if station_id == "GUC:M1":
            LOG.debug(f"Using known GUC M1 location for {station_id}")
            return (38.956158, -106.987856, 2886.0 * 3.28084)

        elif station_id == "GUC:S1":
            LOG.debug(f"Using known GUC S1 location for {station_id}")
            return (38.956158, -106.987856, 2886.0 * 3.28084)

        elif station_id == "GUC:S2":
            LOG.debug(f"Using known GUC S2 location for {station_id}")
            return (38.898361, -106.94314, 3137.0 * 3.28084)

        elif station_id == "GUC:S3":
            LOG.debug(f"Using known GUC S3 location for {station_id}")
            return (38.941556, -106.973128, 2857.0 * 3.28084)

        elif station_id == "GUC:S4":
            LOG.debug(f"Using known GUC S4 location for {station_id}")
            return (38.922019, -106.9509, 2764.0 * 3.28084)
        else:
            LOG.warning(f"Unexpected site information, attmpting to retrieve location for {station_id}")
            if variable is None:
                raise ValueError("Variable must be provided to get location")
            loc = arm_utils.get_station_location(
                site=variable.extra["site"],
                measurement=variable.extra["measurement"],
                facility_code=variable.extra["facility_code"],
                data_level=variable.extra["data_level"],
            )
            return loc



"""Console script for metloom."""
import argparse
import sys
import geopandas as gpd

from metloom.pointdata import CDECPointData, SnotelPointData


def main():
    """Console script for metloom."""

    datasouce_map = {
        "cdec": CDECPointData,
        "snotel": SnotelPointData
    }
    parser = argparse.ArgumentParser(
        description="Find measurement locations within a shapefile"
    )
    parser.add_argument(
        "--shapefile", "-sf", dest="shapefile", required=True,
        help="shapefile used to filter measurement locations"
    )
    parser.add_argument(
        "--datasource", "-ds", dest="datasource",
        choices=list(datasouce_map.keys()),
        required=True, help="datasource used to find measurement ids"
    )
    parser.add_argument(
        "--variables", "-v", dest="variables", nargs="+", required=True,
        help="measurement must posses at least one of these variable. "
             "Variables must be the attribute name of the variables"
             "corresponding to the datasource. i.e. SWE"
    )

    parser.add_argument(
        "--snowcourse", "-sc", dest="snowcourse", default=False,
        action="store_true", required=False,
        help="Do you want to find only snowcourses"
    )

    parser.add_argument(
        "--ouput", "-o", dest="output", default="./output.csv",
        required=False, help="output csv file path"
    )

    args = parser.parse_args()
    datasource = datasouce_map.get(args.datasource)
    if datasource is None:
        raise ValueError("Not a valid datasource")
    geometry = gpd.read_file(args.shapefile)

    variables = []
    for variable_string in args.variables:
        variable = getattr(datasource.ALLOWED_VARIABLES, variable_string, None)
        if variable is None:
            raise ValueError(
                f"Could not find variable corresponding to {variable_string}"
            )
        variables.append(variable)

    points = datasource.points_from_geometry(
        geometry, variables, snow_courses=args.snowcourse
    )

    points.to_dataframe().to_csv(args.output)


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



from logging import getLogger
from typing import Optional
import geopandas as gpd
import pandas as pd
from shapely.geometry import Polygon

from .variables import SensorDescription

LOG = getLogger("metloom.dataframe_utils")


def join_df(
    df: Optional[pd.DataFrame], new_df: Optional[pd.DataFrame], how="left",
    on=None, filter_unused=False
):
    """
    join two dataframes handling None
    Args:
        df: optional dataframe
        new_df: optional dataframe
        how: method for merging
        on: optional kwarg for DataFrame.join
        filter_unused: boolean, whether to filter out columns with _unused in
            then name

    Returns:
        The joined dataframes. This method prefers values from the first if
        columns are overlapping and renames the overlapping values from
        the `new_df` to <column>_unused
    """
    if df is None:
        result_df = new_df
    elif new_df is None:
        result_df = df
    else:
        try:
            result_df = df.join(new_df, how=how, on=on, rsuffix="_unused")
            if filter_unused:
                columns = result_df.columns
                final_columns = [c for c in columns if "_unused" not in c]
                result_df = result_df.filter(final_columns)
        except Exception as e:
            LOG.error("failed joining dataframes.")
            raise e

    return result_df


def merge_df(
    df: Optional[pd.DataFrame], new_df: Optional[pd.DataFrame], how="left"
):
    """
    join two dataframes. Assumes the dataframes are indexed on datetime
    Args:
        df: optional dataframe
        new_df: optional dataframe
    Returns:
        The merged dataframe
    """
    if df is None or len(df) == 0:
        result_df = new_df
    elif new_df is None or len(new_df) == 0:
        result_df = df
    else:
        try:
            result_df = pd.merge_ordered(
                df.reset_index(),
                new_df.reset_index().drop_duplicates()
            )
            result_df.set_index("datetime", inplace=True)
            result_df.sort_index(inplace=True)
            if len(result_df.index.unique()) != len(result_df.index):
                LOG.error("Merging did not result in unique indexes. Killing"
                          " to avoid missing data")
                raise ValueError("Issue merging")
        except Exception as e:
            LOG.error("failed joining dataframes.")
            raise e

    return result_df


def append_df(df: Optional[pd.DataFrame], new_df: Optional[pd.DataFrame]):
    """
    append 2 dfs handling Nones
    Args:
        df: optional dataframe
        new_df: optional dataframe
    Returns:
        dataframe or None
    """
    if df is None:
        result_df = new_df
    elif new_df is None:
        result_df = df
    else:
        result_df = pd.concat([df, new_df])
    return result_df


def resample_df(raw_df: pd.DataFrame,
                variable: SensorDescription, interval: str = 'H'):
    """
    Resample an datatime indexed pandas dateframe to hourly or daily timer
    intervals.
    Resample a datetime indexed pandas dataframe for 1 variable

    Args:
        raw_df: Pandas Dataframe containing a datetime index at an interval
            smaller than hourly.
        variable: SensorDescriptions to be found in the dataframe
        interval: Interval to resample to. Options are H = Hourly, D=Daily

    Returns:
        df: Pandas Dataframe of a single variable resampled to the
            desired interval
    """
    name = variable.name
    if name in raw_df.columns:
        if variable.accumulated:
            result = raw_df[name].resample(interval).sum()
        else:
            result = raw_df[name].resample(interval).mean()
        df = pd.DataFrame()
        df[name] = result
        df = df.dropna()
    else:
        df = None

    return df


def resample_series(raw_series: pd.Series, variable: SensorDescription, interval: str = "h"):
    """
    Resample a pandas series to hourly or daily timer intervals.
    Resample a datetime indexed pandas series for 1 variable

    Args:
        raw_series: Pandas Series containing a datetime index at an interval
            smaller than hourly.
        variable: SensorDescriptions to be found in the dataframe
        interval: Interval to resample to. Options are H = Hourly, D=Daily

    Returns:
        Pandas Series of a single variable resampled to the desired interval
    """
    if variable.accumulated:
        result = raw_series.resample(interval).sum()
    else:
        result = raw_series.resample(interval).mean()

    return result


def resample_whole_df(raw_df: pd.DataFrame, variable: SensorDescription, interval: str = "h"):
    """
    Resample an datatime indexed pandas dateframe to hourly or daily timer
    intervals.
    Resample a datetime indexed pandas dataframe for 1 variable
    Other columns get resampled with the .first() values

    Args:
        raw_df: Pandas Dataframe containing a datetime index at an interval
            smaller than hourly.
        variable: SensorDescriptions to be found in the dataframe
        interval: Interval to resample to. Options are H = Hourly, D=Daily

    Returns:
        df: Pandas Dataframe of a single variable resampled to the
            desired interval
    """
    name = variable.name
    df = pd.DataFrame()
    columns = raw_df.columns
    if name in raw_df.columns:
        if variable.accumulated:
            result = raw_df[name].resample(interval).sum()
        else:
            result = raw_df[name].resample(interval).mean()
        df[name] = result
        df = df.dropna()
        # get the first value for the other columns
        for c in columns:
            if c != name:
                df[c] = raw_df[c].resample(interval).first()
    else:
        df = None

    return df


def shp_to_box(geometry):
    """
    Convert a shapefiles geodataframe to the bounding box of the shapefile
    as a new geodataframe

    Args:
        geometry: geodataframe polygon

    Returns:
        geodataframe of the bounding box
    """
    bounds = geometry.total_bounds  # Returns a tuple (minx, miny, maxx, maxy)

    # Create a Polygon from the bounding box coordinates
    bounding_box = Polygon([
        (bounds[0], bounds[1]),
        (bounds[0], bounds[3]),
        (bounds[2], bounds[3]),
        (bounds[2], bounds[1])
    ])

    # Create a new GeoDataFrame to hold the bounding box geometry
    gdf_bbox = gpd.GeoDataFrame({'geometry': [bounding_box]}, crs=geometry.crs)
    return gdf_bbox



"""
Utility functions for U.S. Department of Energy’s Atmospheric Radiation Measurement (ARM) User Facility.
https://www.arm.gov/

The download tools here will modelled after information from the ARM Live Data Webservice
https://adc.arm.gov/armlive/register#web_services
"""

import os
import re
import concurrent.futures
import pydash
import logging
import pathlib
import json
import requests
import pandas as pd
import typing
from datetime import date, datetime

LOG = logging.getLogger(__name__)
ARM_FILE_DATE_REGEX = re.compile(r".*?\.(?P<date>\d{8})\.\.*")


def get_station_data(
    *,
    site: str,
    measurement: str,
    facility_code: str,
    data_level: str,
    variables: list[str],
    start: typing.Union[date, datetime] = None,
    end: typing.Union[date, datetime] = None,
    token_json: pathlib.Path = pathlib.Path("~/.arm_token.json").expanduser(),
    access_token=None,
    user_id=None,
    destination: pathlib.Path = pathlib.Path(".armdata"),
    download_chunk_size: int = 20,
    url: str = "https://adc.arm.gov/armlive/data",
) -> pd.DataFrame:
    """
    Get data from the ARM Live Data Webservice as a pandas DataFrame.

    This function queries the ARM Live Data Webservice with the provided parameters,
    downloads the resulting files, and returns their contents as a pandas DataFrame.

    Args:
        site: Site identifier code (e.g., 'sgp', 'nsa', etc.)
        measurement: Measurement identifier code.
        facility_code: Facility identifier code.
        data_level: Data processing level (e.g., 'c1', 'b1', etc.)
        variables: List of variable names to download.
        start: Start date/datetime. Defaults to None, which downloads all available data.
        end: End date/datetime. Defaults to None, which downloads all available data.
        token_json: Path to a JSON file containing the access token and user ID. Defaults to ~/.arm_token.json.
                    If provided, it will take precedence over the environment variables for access_token and user_id.
        access_token: ARM access token. Defaults to M3W_ARM_ACCESS_TOKEN environment variable.
        user_id: ARM user ID. Defaults to M3W_ARM_USER_ID environment variable.
        destination: Directory to save downloaded files. Defaults to ".armdata".
        download_chunk_size: Number of files to download in each batch. Defaults to 20.
        url: Base URL for the ARM Live Data Webservice. Defaults to "https://adc.arm.gov/armlive/data".

    Returns:
        pd.DataFrame: DataFrame containing the requested variables with datetime index.

    Raises:
        AssertionError: If user_id or access_token is not provided.
        HTTPError: If the API request fails.
    """
    if token_json.is_file():
        LOG.info(f"Loading token from {token_json}")
        with open(token_json, "r") as f:
            token = json.load(f)
        if access_token is None:
            access_token = token.get("access_token", None)
        if user_id is None:
            user_id = token.get("user_id", None)

    # set the default values for access_token and user_id
    access_token = access_token or os.getenv("M3W_ARM_ACCESS_TOKEN", None)
    user_id = user_id or os.getenv("M3W_ARM_USER_ID", None)

    # check if user_id and access_token are provided
    if user_id is None:
        raise ValueError(
            "user_id must be provided, set M3W_ARM_USER_ID environment variable or token_json JSON file.")
    if access_token is None:
        raise ValueError(
            "access_token must be provided, set M3W_ARM_ACCESS_TOKEN environment variable or token_json JSON file.")

    #  define the parameters for the query
    q_params = dict(
        user=f"{user_id}:{access_token}",
        ds=(
            f"{site.lower()}{measurement.lower()}{facility_code.upper()}.{data_level.lower()}"
        ),
        wt="json",
    )
    if start is not None:
        start = datetime.fromisoformat(start) if isinstance(start, str) else start
        q_params["start"] = (
            start.date().isoformat() if hasattr(start, "date") else start.isoformat()
        )
    if end is not None:
        end = datetime.fromisoformat(end) if isinstance(end, str) else end
        q_params["end"] = (
            end.date().isoformat() if hasattr(end, "date") else end.isoformat()
        )

    # make the request to the ARM Live Data Webservice
    LOG.info(f"Querying {url}/query for data.")
    response = requests.get(f"{url}/query", params=q_params)
    response.raise_for_status()

    # extract the files from the response
    files = sorted(response.json().get("files", []))
    if not files:
        LOG.warning(
            "No files found for the given parameters, check the values and case is correct for site, measurement, "
            "facility_code, and data_level."
        )
        return

    # download the data, with the desired variables as a single csv file
    # the ARM API allows for multiple files to be download, but when I attemped to do all of them at once it failed
    LOG.info(f"Downloading {len(files)} files, this may take a several minutes.")
    destination.mkdir(parents=True, exist_ok=True)
    with concurrent.futures.ThreadPoolExecutor() as executor:
        jobs = set()
        for i, chunk in enumerate(pydash.chunk(files, download_chunk_size)):
            jobs.add(
                executor.submit(
                    _download_arm_files,
                    user_id=user_id,
                    access_token=access_token,
                    ds=q_params["ds"],
                    url=url,
                    variables=variables,
                    destination=destination,
                    files=chunk,
                )
            )
        csv_files = [
            future.result() for future in concurrent.futures.as_completed(jobs)
        ]

    # read the csv files into a single DataFrame
    LOG.info("Reading files into a single DataFrame.")
    df = pd.concat(
        [pd.read_csv(f, index_col="time", parse_dates=True) for f in csv_files],
        axis="index",
    )
    df.sort_index(inplace=True)
    df = df[~df.index.duplicated(keep="first")]
    df.index.name = "datetime"
    return df


def get_station_location(
    site: str,
    measurement: str,
    facility_code: str,
    data_level: str,
    start: typing.Union[date, datetime] = None,
    end: typing.Union[date, datetime] = None,
) -> typing.Tuple[float, float, float]:
    """
    Get the location of a station from the ARM Live Data Webservice.
    """

    # download the data for the given site, measurement, facility_code, and data_level
    df = get_station_data(
        site=site,
        measurement=measurement,
        facility_code=facility_code,
        data_level=data_level,
        start=start,
        end=end,
        variables=["lat", "lon", "alt"],
    )

    # it is possible that the location is not unique, so we need to check for that
    # for example if the data is from a site with a balloon sensor
    lat = _get_location_helper(df["lat"], "latitude")
    lon = _get_location_helper(df["lon"], "longitude")
    alt = _get_location_helper(df["alt"], "altitude")
    return lat, lon, alt * 3.28084  # convert meters to feet


def _get_location_helper(data: pd.Series, text: str) -> float:
    """
    Check if the data has non-unique values and log a warning if so.
    """
    unique_data = data.drop_duplicates()
    if len(unique_data) > 1:
        LOG.warning(
            f"Multiple {text} found, the mean is being provided: mean={data.mean()}, min={data.min()}, max={data.max()}"
        )
        return float(data.mean())
    return float(unique_data.iloc[0])


def _download_arm_files(
    user_id: str,
    access_token: str,
    ds: str,
    url: str,
    variables: list[str],
    destination: os.PathLike,
    files: list[str],
) -> list[os.PathLike]:
    """
    Download a list of files from the ARM Live Data Webservice, using /saveData
    """
    d0 = ARM_FILE_DATE_REGEX.search(files[0]).group("date")
    dn = ARM_FILE_DATE_REGEX.search(files[-1]).group("date")
    dv = "_".join(variables)
    output = pathlib.Path(destination, f"{ds}_{dv}_{d0}_{dn}.csv")

    if output.exists():
        LOG.debug(f"File {output} already exists, skipping download.")
        return output

    params = dict(
        user=f"{user_id}:{access_token}", variables=",".join(variables), wt="csv"
    )
    response = requests.get(f"{url}/mod", params=params, data=json.dumps(files))
    response.raise_for_status()
    with open(output, "wb") as f:
        f.write(response.content)

    return output



"""
Taken from
stackoverflow.com/questions/15445981/how-do-i-disable-the-security-certificate-check-in-python-requests
This is necessary to make requests to Snotel without exceptions
as we are not able to do the SSL verification
"""
import warnings
import contextlib
import requests
from urllib3.exceptions import InsecureRequestWarning

old_merge_environment_settings = requests.Session.merge_environment_settings


@contextlib.contextmanager
def no_ssl_verification():
    opened_adapters = set()

    def merge_environment_settings(self, url, proxies, stream, verify, cert):
        # Verification happens only once per connection so we need to close
        # all the opened adapters once we're done. Otherwise, the effects of
        # verify=False persist beyond the end of this context manager.
        opened_adapters.add(self.get_adapter(url))

        settings = old_merge_environment_settings(
            self, url, proxies, stream, verify, cert
        )
        settings['verify'] = False

        return settings

    requests.Session.merge_environment_settings = merge_environment_settings

    try:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', InsecureRequestWarning)
            yield
    finally:
        requests.Session.merge_environment_settings = old_merge_environment_settings

        for adapter in opened_adapters:
            try:
                adapter.close()
            except Exception:
                pass



import pytest
import geopandas as gpd
import pandas as pd
from os import path

from metloom.pointdata.base import PointData, DataValidationError


def side_effect_error(*args):
    raise ValueError("Testing error")


class TestPointData:
    def test_class_attributes(self):
        # Base implementation should fail
        with pytest.raises(AttributeError):
            PointData("foo", "bar").tzinfo


class BasePointDataTest(object):
    @pytest.fixture(scope="class")
    def data_dir(self):
        this_dir = path.dirname(__file__)
        return path.join(this_dir, "data")

    @pytest.fixture(scope="class")
    def shape_obj(self, data_dir):
        fp = path.join(data_dir, "testing.shp")
        return gpd.read_file(fp)

    @staticmethod
    def expected_response(dates, variables_map, station, points,
                          include_measurement_date=False):
        obj = []
        for idt, dt in enumerate(dates):
            # get the value and unit corresponding to the date
            row_obj = {k: v[idt] for k, v in variables_map.items()}
            entry = {
                "datetime": pd.Timestamp(dt, tz="UTC"),
                "site": station.id,
                "datasource": "NRCS",
                **row_obj
            }
            if include_measurement_date:
                entry["measurementDate"] = pd.Timestamp(dt, tz="UTC")
            obj.append(
                entry
            )
        df = gpd.GeoDataFrame.from_dict(
            obj,
            geometry=[points] * len(dates),
        )
        # needed to reorder the columns for the pd testing compare
        var_keys = list(variables_map.keys())
        var_keys.sort()
        df = df.filter(
            [
                "datetime", "geometry", "site", "measurementDate",
                *var_keys, "datasource"]
        )
        df.set_index(keys=["datetime", "site"], inplace=True)
        return df.sort_index(axis=1)


class TestPointDataValidations:
    def test_validate_gdf(self):
        with pytest.raises(DataValidationError):
            PointData._validate_geodataframe(pd.DataFrame())

    def test_validate_indicies(self):
        with pytest.raises(DataValidationError):
            PointData._validate_df_indicies(gpd.GeoDataFrame(index=["datetime"]))

    @pytest.mark.parametrize("incoming_columns", [
        # Missing datasource
        ["geometry", "measurementDate", "quality_code"],
        ["datasource"],
    ])
    def test_validate_columns(self, incoming_columns):
        gdf = gpd.GeoDataFrame(columns=incoming_columns)
        with pytest.raises(DataValidationError):
            columns = PointData.EXPECTED_COLUMNS.copy()
            PointData._validate_df_columns(gdf, columns)

    def test_validate_units(self):
        # Missing units for this variables
        incoming = ["Depth", 'SWE']
        gdf = gpd.GeoDataFrame(columns=incoming)
        with pytest.raises(DataValidationError):
            PointData._validate_df_units(gdf, PointData.EXPECTED_COLUMNS)



from .base import PointData, PointDataCollection
from .cdec import CDECPointData
from .snotel import SnotelPointData
from .mesowest import MesowestPointData
from .usgs import USGSPointData
from .geosphere_austria import GeoSphereHistPointData, GeoSphereCurrentPointData
from .norway import MetNorwayPointData
from .cues import CuesLevel1
from .nws_forecast import NWSForecastPointData
from .files import CSVPointData, StationInfo
from .snowex import SnowExMet
from .csas import CSASMet
from .sail import SAILPointData

__all__ = [
    "PointData", "PointDataCollection", "CDECPointData", "SnotelPointData",
    "MesowestPointData", "USGSPointData", "GeoSphereHistPointData",
    "GeoSphereCurrentPointData", "CuesLevel1", "MetNorwayPointData",
    "NWSForecastPointData",
    "CSVPointData", "StationInfo", "SnowExMet", "CSASMet", "SAILPointData"
]



from datetime import timezone, timedelta, datetime
from unittest.mock import patch, MagicMock
from collections import OrderedDict

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest

from metloom.pointdata import SnotelPointData
from metloom.variables import SnotelVariables
from tests.test_point_data import BasePointDataTest


class MockZeepObject:
    def __init__(self, obj):
        """
        Args:
            obj: dictionary of values
        """
        self.mock_dict = OrderedDict(obj)
        self.__values__ = self.mock_dict
        for k, v in obj.items():
            setattr(self, k, v)

    def __dict__(self):
        return self.mock_dict

    def __getitem__(self, item):
        return self.mock_dict[item]


class TestSnotelPointData(BasePointDataTest):

    @pytest.fixture(scope="class")
    def points(self):
        return gpd.points_from_xy([-107.67552], [37.9339], z=[9800.0])[0]

    @staticmethod
    def snotel_meta_sideeffect(*args, **kwargs):
        """
        Mock out the metadata response
        """
        code = kwargs["stationTriplet"]
        available_stations = {
            "538:CO:SNTL": {
                "actonId": "07M27S",
                "beginDate": "1979-10-01 00:00:00",
                "countyName": "Ouray",
                "elevation": 9800.0,
                "endDate": "2100-01-01 00:00:00",
                "fipsCountryCd": "US",
                "fipsCountyCd": "091",
                "fipsStateNumber": "08",
                "huc": "140200060201",
                "hud": "14020006",
                "latitude": 37.9339,
                "longitude": -107.67552,
                "name": "Idarado",
                "shefId": "IDRC2",
                "stationTriplet": "538:CO:SNTL",
                "stationDataTimeZone": -8.0
            },
            "538:CO:SNOW": {
                "actonId": "07M27S",
                "beginDate": "1979-10-01 00:00:00",
                "countyName": "Ouray",
                "elevation": 9800.0,
                "endDate": "2100-01-01 00:00:00",
                "fipsCountryCd": "US",
                "fipsCountyCd": "091",
                "fipsStateNumber": "08",
                "huc": "140200060201",
                "hud": "14020006",
                "latitude": 37.9339,
                "longitude": -107.67552,
                "name": "Idarado",
                "shefId": "IDRC2",
                "stationTriplet": "538:CO:SNOW",
            },
            "FFF:CA:SNOW": {
                "actonId": None,
                "beginDate": "1930-02-01 00:00:00",
                "countyName": "Tuolumne",
                "elevation": 6500.0,
                "endDate": "2100-01-01 00:00:00",
                "fipsCountryCd": "US",
                "fipsCountyCd": "109",
                "fipsStateNumber": "06",
                "huc": "180400090302",
                "latitude": 37.995,
                "longitude": -119.78,
                "name": "Fake1",
                "shefId": None,
                "stationTriplet": "FFF:CA:SNOW",
            },
            "BBB:CA:SNOW": {
                "actonId": None,
                "beginDate": "1948-02-01 00:00:00",
                "countyName": "Tuolumne",
                "elevation": 9300.0,
                "endDate": "2100-01-01 00:00:00",
                "fipsCountryCd": "US",
                "fipsCountyCd": "109",
                "fipsStateNumber": "06",
                "huc": "180400090402",
                "hud": "18040009",
                "latitude": 38.18333,
                "longitude": -119.61667,
                "name": "Fake2",
                "shefId": None,
                "stationTriplet": "BBB:CA:SNOW",
            },
        }
        return MockZeepObject(available_stations[code])

    @staticmethod
    def snotel_data_sideeffect(*args, **kwargs):
        duration = kwargs["duration"]
        if duration == "SEMIMONTHLY":
            return [
                MockZeepObject({
                    'beginDate': '2020-01-20 00:00:00',
                    'collectionDates': ['2020-01-28', '2020-02-27'],
                    'duration': 'SEMIMONTHLY',
                    'endDate': '2020-03-14 00:00:00',
                    'flags': ['V', 'V'], 'stationTriplet': '538:CO:SNTL',
                    'values': [13.19, 13.17]})
            ]
        if duration == "DAILY":
            return [MockZeepObject({
                'beginDate': '2020-03-20 00:00:00',
                'collectionDates': [], 'duration': 'DAILY',
                'endDate': '2020-03-22 00:00:00', 'flags': ['V', 'V', 'V'],
                'stationTriplet': '538:CO:SNTL',
                'values': [13.19, 13.17, 13.14]})]

    @staticmethod
    def snotel_hourly_sideeffect(*args, **kwargs):
        element_cd = kwargs["elementCd"]
        if element_cd == "WTEQ":
            return [
                {
                    'beginDate': '2020-01-02 00:00', 'endDate': '2020-01-20 00:00',
                    'stationTriplet': '538:CO:SNTL',
                    'values': [
                        {
                            'dateTime': '2020-03-20 00:00',
                            'flag': 'V',
                            'value': 13.19
                        }, {
                            'dateTime': '2020-03-20 01:00',
                            'flag': 'V',
                            'value': 13.17
                        }, {
                            'dateTime': '2020-03-20 02:00',
                            'flag': 'V',
                            'value': 13.14
                        }]}]
        elif element_cd == "PRCPSA":
            return [
                {
                    'beginDate': '2020-01-02 00:00',
                    'endDate': '2020-01-20 00:00',
                    'stationTriplet': '538:CO:SNTL',
                    'values': [
                        {
                            'dateTime': '2020-03-20 00:00',
                            'flag': 'V',
                            'value': 4.1
                        }, {
                            'dateTime': '2020-03-20 02:00',
                            'flag': 'V',
                            'value': 4.3
                        }, {
                            'dateTime': '2020-03-20 03:00',
                            'flag': 'V',
                            'value': 4.4
                        }]}]
        elif element_cd == "STO":
            return [
                {
                    'beginDate': '2020-01-02 00:00',
                    'endDate': '2020-01-20 00:00',
                    'stationTriplet': '538:CO:SNTL',
                    'values': [
                        {
                            'dateTime': '2020-03-20 00:00',
                            'flag': 'V',
                            'value': -0.3,
                        }, {
                            'dateTime': '2020-03-20 01:00',
                            'flag': 'V',
                            'value': -0.4,
                        }, {
                            'dateTime': '2020-03-20 02:00',
                            'flag': 'V',
                            'value': -0.5,
                        }]}]
        else:
            raise ValueError(f"{element_cd} not configured in this mock")

    @pytest.fixture(scope="class")
    def mock_elements(self):
        return [
            MockZeepObject(
                {
                    'beginDate': '1980-07-23 00:00:00', 'dataPrecision': 1,
                    'duration': 'DAILY', 'elementCd': 'WTEQ',
                    'endDate': '2100-01-01 00:00:00', 'heightDepth': None,
                    'ordinal': 1,
                    'originalUnitCd': 'in', 'stationTriplet': '538:CO:SNTL',
                    'storedUnitCd': 'in'}),
            MockZeepObject(
                {'beginDate': '1980-07-23 00:00:00', 'dataPrecision': 1,
                 'duration': 'SEMIMONTHLY', 'elementCd': 'WTEQ',
                 'endDate': '2100-01-01 00:00:00', 'heightDepth': None,
                 'ordinal': 1,
                 'originalUnitCd': 'in', 'stationTriplet': '538:CO:SNTL',
                 'storedUnitCd': 'in'}),
            MockZeepObject(
                {'beginDate': '1979-10-01 00:00:00', 'dataPrecision': 1,
                 'duration': 'HOURLY', 'elementCd': 'WTEQ',
                 'endDate': '2100-01-01 00:00:00', 'heightDepth': None,
                 'ordinal': 1,
                 'originalUnitCd': 'in', 'stationTriplet': '538:CO:SNTL',
                 'storedUnitCd': 'in'}),
            MockZeepObject(
                {'beginDate': '1980-07-23 00:00:00', 'dataPrecision': 1,
                 'duration': 'MONTHLY', 'elementCd': 'WTEQ',
                 'endDate': '2100-01-01 00:00:00', 'heightDepth': None,
                 'ordinal': 1,
                 'originalUnitCd': 'in', 'stationTriplet': '538:CO:SNTL',
                 'storedUnitCd': 'in'}),
            MockZeepObject(
                {'beginDate': '1979-10-01 00:00:00', 'dataPrecision': 1,
                 'duration': 'HOURLY', 'elementCd': 'PRCPSA',
                 'endDate': '2100-01-01 00:00:00', 'heightDepth': None,
                 'ordinal': 1,
                 'originalUnitCd': 'in', 'stationTriplet': '538:CO:SNTL',
                 'storedUnitCd': 'in'}),
            MockZeepObject(
                {'beginDate': '1979-10-01 00:00:00', 'dataPrecision': 1,
                 'duration': 'HOURLY', 'elementCd': 'STO',
                 'endDate': '2100-01-01 00:00:00',
                 'heightDepth': {'unitCd': 'in', 'value': '-2'},
                 'ordinal': 1,
                 'originalUnitCd': 'degF', 'stationTriplet': '538:CO:SNTL',
                 'storedUnitCd': 'degF'})
        ]

    @pytest.fixture
    def mock_zeep_client(self, mock_elements):
        with patch("metloom.pointdata.snotel_client.zeep.Client") as mock_client:
            mock_service = MagicMock()
            # setup the individual services
            mock_service.getStationMetadata.side_effect = self.snotel_meta_sideeffect
            mock_service.getStationElements.return_value = mock_elements
            mock_service.getStations.return_value = ["FFF:CA:SNOW",
                                                     "BBB:CA:SNOW"]
            mock_service.getData.side_effect = self.snotel_data_sideeffect
            mock_service.getHourlyData.side_effect = self.snotel_hourly_sideeffect
            # assign service to client
            mock_client.return_value.service = mock_service
            yield mock_client

    def test_metadata(self, mock_zeep_client):
        obj = SnotelPointData("538:CO:SNTL", "eh")
        assert (
            obj.metadata == gpd.points_from_xy(
                [-107.67552], [37.9339], z=[9800.0])[0]
        )
        assert obj.tzinfo == timezone(timedelta(hours=-8.0))

    @pytest.mark.parametrize(
        "station_id, dts, expected_dts, vals, d1, d2, fn_name",
        [
            (
                "538:CO:SNTL",
                ["2020-03-20 00:00", "2020-03-20 01:00", "2020-03-20 02:00"],
                ["2020-03-20 08:00", "2020-03-20 09:00", "2020-03-20 10:00"],
                {
                    SnotelVariables.SWE.name: [13.19, 13.17, 13.14],
                    f"{SnotelVariables.SWE.name}_units": ["in", "in", "in"]
                },
                datetime(2020, 3, 20, 0),
                datetime(2020, 3, 20, 2),
                "get_hourly_data",
            ),
            (
                "538:CO:SNTL",
                ["2020-03-20 00:00", "2020-03-20 01:00", "2020-03-20 02:00"],
                ["2020-03-20 08:00", "2020-03-20 09:00", "2020-03-20 10:00"],
                {
                    SnotelVariables.TEMPGROUND2IN.name: [-0.3, -0.4, -0.5],
                    f"{SnotelVariables.TEMPGROUND2IN.name}_units":
                        ["degF", "degF", "degF"]
                },
                datetime(2020, 3, 20, 0),
                datetime(2020, 3, 20, 2),
                "get_hourly_data",
            ),
            (
                "538:CO:SNTL",
                ["2020-03-20", "2020-03-21", "2020-03-22"],
                ["2020-03-20 08:00", "2020-03-21 08:00", "2020-03-22 08:00"],
                {
                    SnotelVariables.SWE.name: [13.19, 13.17, 13.14],
                    f"{SnotelVariables.SWE.name}_units": ["in", "in", "in"]
                },
                datetime(2020, 3, 20),
                datetime(2020, 3, 22),
                "get_daily_data",
            ),
            (
                "538:CO:SNOW",
                ["2020-01-28", "2020-02-27"],
                ["2020-01-28 00:00", "2020-02-27 00:00"],
                {
                    SnotelVariables.SWE.name: [13.19, 13.17],
                    f"{SnotelVariables.SWE.name}_units": ["in", "in"]
                },
                datetime(2020, 1, 20),
                datetime(2020, 3, 15),
                "get_snow_course_data",
            ),
        ],
    )
    def test_get_data_methods(
            self, station_id, dts, expected_dts, vals, d1,
            d2, fn_name, points, mock_zeep_client):
        station = SnotelPointData(station_id, "TestSite")
        if 'GROUND TEMPERATURE -2IN' in list(vals.keys()):
            vrs = [SnotelVariables.TEMPGROUND2IN]
        else:
            vrs = [SnotelVariables.SWE]
        fn = getattr(station, fn_name)
        result = fn(d1, d2, vrs)
        expected = self.expected_response(
            expected_dts, vals, station, points,
            include_measurement_date="snow_course" in fn_name
        )
        pd.testing.assert_frame_equal(
            result.sort_index(axis=1), expected
        )

    def test_get_hourly_data_multi_sensor(self, points, mock_zeep_client):
        expected_dts = [
            "2020-03-20 08:00", "2020-03-20 09:00", "2020-03-20 10:00",
            "2020-03-20 11:00"
        ]
        expected_vals_obj = {
            SnotelVariables.SWE.name: [13.19, 13.17, 13.14, np.nan],
            f"{SnotelVariables.SWE.name}_units": ["in", "in", "in", np.nan],
            SnotelVariables.PRECIPITATION.name: [4.1, np.nan, 4.3, 4.4],
            f"{SnotelVariables.PRECIPITATION.name}_units": [
                "in", np.nan, "in", "in"],
        }
        station = SnotelPointData("538:CO:SNTL", "TestSite")
        vrs = [SnotelVariables.PRECIPITATION, SnotelVariables.SWE]
        result = station.get_hourly_data(
            datetime(2020, 3, 20, 0), datetime(2020, 3, 20, 4), vrs
        )
        expected = self.expected_response(
            expected_dts, expected_vals_obj, station, points
        )
        pd.testing.assert_frame_equal(
            result.sort_index(axis=1), expected
        )

    def test_points_from_geometry(self, shape_obj, mock_zeep_client):
        result = SnotelPointData.points_from_geometry(
            shape_obj, [SnotelVariables.SWE], snow_courses=True
        )
        ids = [point.id for point in result]
        names = [point.name for point in result]
        assert len(names) == 2
        assert set(ids) == {"FFF:CA:SNOW", "BBB:CA:SNOW"}
        assert set(names) == {"Fake1", "Fake2"}

    def test_points_from_geomtery_buffer(self, shape_obj, mock_zeep_client):
        SnotelPointData.points_from_geometry(
            shape_obj, [SnotelVariables.SWE], snow_courses=False, buffer=0.1
        )
        search_kwargs = mock_zeep_client().method_calls[0][2]
        expected = {
            'maxLatitude': 38.3, 'minLatitude': 37.6,
            'maxLongitude': -119.1, 'minLongitude': -119.9
        }
        for k, v in expected.items():
            assert v == pytest.approx(search_kwargs[k])

    def test_points_from_geometry_fail(self, shape_obj, mock_zeep_client):
        mock_zeep_client.return_value.service.getStations.return_value = []
        result = SnotelPointData.points_from_geometry(
            shape_obj, [SnotelVariables.SWE], snow_courses=True
        )
        assert len(result) == 0



"""
Data reader for the Center for Snow and Avalanche Studies
"""
from metloom.pointdata import CSVPointData, StationInfo
from metloom.variables import CSASVariables
import os
from datetime import datetime, timedelta


class InvalidDateRange(Exception):
    """
    Exception to indicate there is no know data for the available date range
    """


class CSASStationInfo(StationInfo):
    # Name, id, lat, long, elevation, http path
    SENATOR_BECK = ("Senator Beck Study Plot", "SBSP", 37.90688, -107.72627, 12186,
                    "2023/11/SBSP_1hr_2003-2009.csv")
    SWAMP_ANGEL = ("Swamp Angel Study Plot", "SASP", 37.90691, -107.71132, 11060,
                   "2023/11/SASP_1hr_2003-2009.csv")
    PUTNEY = ("Putney Study Plot", "PTSP", 37.89233, -107.69577, 12323,
              "2025/05/PTSP_1hr.csv")
    SENATOR_BECK_STREAM_GAUGE = ("Senator Beck Stream Gauge", "SBSG", 37.90678,
                                 -107.70943, 11030, "2025/05/SBSG_1hr.csv")


class CSASMet(CSVPointData):
    """
    """
    CURRENT_AVAILABLE_YEAR = 2024
    ALLOWED_VARIABLES = CSASVariables
    ALLOWED_STATIONS = CSASStationInfo

    # Data is in Mountain time
    UTC_OFFSET_HOURS = -7

    URL = "https://snowstudies.org/wp-content/uploads/"
    DATASOURCE = "CSAS"
    DOI = ""

    def _file_urls(self, station_id, start, end):
        """
        Navigate the system using dates. Data for SASP and SBSP is stored in
        two csvs. 2003-2009 and 2010-2023. Not sure what happens when the
        next year is made available. This function will grab the necessary urls
        depending on the requested data
        """
        urls = []

        if station_id in ['SASP', 'SBSP']:
            current_available_year = self.CURRENT_AVAILABLE_YEAR

            if start.year <= 2009:
                urls.append(os.path.join(self.URL, self._station_info.path))

            # Account for later file use or even straddling thge data
            if start.year > 2009 or end.year > 2009:  # TODO: add to the info enum?
                partial = str(self._station_info.path).replace("2003", "2010")
                partial = partial.replace('2009', str(current_available_year))
                # Change the subdirectory in the url
                filename = partial.replace('2023/11/', '2025/05/')

                urls.append(os.path.join(self.URL, filename))

            if start.year < 2003 or end.year > current_available_year:
                raise InvalidDateRange(f"CSAS data is only available from 2003-"
                                       f"{current_available_year}")
        else:
            urls.append(os.path.join(self.URL, self._station_info.path))

        return urls

    @staticmethod
    def _parse_datetime(row):
        # Julian day is not zero based Jan 1 == DOY 1
        dt = timedelta(days=int(row['DOY']) - 1, hours=int(row['Hour'] / 100))
        return datetime(int(row['Year']), 1, 1) + dt

    def _assign_datetime(self, resp_df):
        resp_df['datetime'] = resp_df.apply(lambda row: self._parse_datetime(row),
                                            axis=1)
        return resp_df.set_index('datetime')



import os
import geopandas as gpd
from shapely.geometry import Polygon
from datetime import datetime
from metloom.pointdata.sail import SAILPointData
from metloom.variables import SAILStationVariables
from unittest import mock
import pytest


@pytest.fixture
def mock_requests_get():
    with mock.patch("requests.get") as mock_get:
        # query
        obj0 = mock.MagicMock()
        obj0.content = None
        obj0.raise_for_status = lambda: None
        obj0.json = lambda: {
            "files": [
                os.path.join(os.path.dirname(__file__), "data", "gucwbpluvio2M1.a1.20230101.000000.nc"),
                os.path.join(os.path.dirname(__file__), "data", "gucwbpluvio2M1.a1.20230102.000000.nc"),
            ]
        }

        # mod
        obj1 = mock.MagicMock()
        with open(
            os.path.join(os.path.dirname(__file__), "data", "gucwbpluvio2M1.a1_accum_rtnrt_20230101_20230102.csv"), "rb"
        ) as f:
            obj1.content = f.read()
        obj1.raise_for_status = lambda: None

        mock_get.side_effect = [obj0, obj1]
        yield mock_get


@pytest.fixture(scope="session")
def setup_env():
    os.environ["M3W_ARM_USER_ID"] = "skroob"
    os.environ["M3W_ARM_ACCESS_TOKEN"] = "12345"


def test_get_hourly_data(mock_requests_get, setup_env):
    # the nc files exist, this test will not download
    obj = SAILPointData("GUC:M1")
    df = obj.get_hourly_data("2023-01-01", "2023-01-02", [SAILStationVariables.PRECIPITATION])
    assert df is not None
    assert len(df) == 48
    assert df.iloc[24]["PRECIPITATION"] == pytest.approx(0.01233333)
    assert df.iloc[24]["PRECIPITATION_units"] == "mm"


def test_get_daily_data(mock_requests_get, setup_env):
    # this mocks the necessary function to test the download code
    obj = SAILPointData(station_id="GUC:M1")
    df = obj.get_daily_data("2023-01-01", "2023-01-02", [SAILStationVariables.PRECIPITATION])
    assert df is not None
    assert len(df) == 2
    assert df.iloc[1]["PRECIPITATION"] == pytest.approx(0.004354167)
    assert df.iloc[1]["PRECIPITATION_units"] == "mm"


def test_check_start_end_dates():
    obj = SAILPointData(station_id="GUC:M1")
    # Test with end_date before start_date
    with pytest.raises(ValueError):
        obj._check_start_end_dates(datetime.fromisoformat("2023-01-02"), datetime.fromisoformat("2023-01-01"))
    # Test with start_date before 2021-09-01
    with pytest.raises(ValueError):
        obj._check_start_end_dates(datetime.fromisoformat("2020-09-01"), datetime.fromisoformat("2023-01-01"))
    # Test with end_date after after 2023-06-16
    with pytest.raises(ValueError):
        obj._check_start_end_dates(datetime.fromisoformat("2023-01-01"), datetime.fromisoformat("2024-01-01"))


@pytest.fixture
def station_inside_gdf():
    geometry = Polygon([(-106.882456, 38.957106), (-107.049037, 38.881546), (-107.036609, 39.004078)])
    return gpd.GeoSeries(geometry, crs="EPSG:4326")


@pytest.fixture
def station_outside_gdf():
    geometry = Polygon([(-106.882456, 38.94), (-107.049037, 38.97), (-107.036609, 39.004078)])
    return gpd.GeoSeries(geometry, crs="EPSG:4326")


def test_points_from_geometry_within_false(station_inside_gdf, station_outside_gdf):
    # bounding box, both should be within bounding box
    gdf = SAILPointData.points_from_geometry(
        geometry=station_inside_gdf,
        variables=[SAILStationVariables.PRECIPITATION],
        within_geometry=False,
        buffer=0.0,
    )
    assert len(gdf) == 1

    gdf = SAILPointData.points_from_geometry(
        geometry=station_outside_gdf,
        variables=[SAILStationVariables.PRECIPITATION],
        within_geometry=False,
        buffer=0.0,
    )
    assert len(gdf) == 1


def test_points_from_geometry_within_true(station_inside_gdf, station_outside_gdf):
    gdf = SAILPointData.points_from_geometry(
        geometry=station_inside_gdf,
        variables=[SAILStationVariables.PRECIPITATION],
        within_geometry=True,
        buffer=0.0,
    )
    assert len(gdf) == 1

    gdf = SAILPointData.points_from_geometry(
        geometry=station_outside_gdf,
        variables=[SAILStationVariables.PRECIPITATION],
        within_geometry=True,
        buffer=0.0,
    )
    assert len(gdf) == 0


def test_points_from_geometry_within_true_with_buffer(station_inside_gdf, station_outside_gdf):
    gdf = SAILPointData.points_from_geometry(
        geometry=station_outside_gdf,
        variables=[SAILStationVariables.PRECIPITATION],
        within_geometry=True,
        buffer=0.3,
    )
    assert len(gdf) == 1



from datetime import datetime, timezone, timedelta
from io import StringIO
from typing import List

import geopandas as gpd
import numpy as np
import pandas as pd
import requests
import logging

from geopandas import GeoDataFrame

from .base import PointData
from ..variables import CdecStationVariables, SensorDescription
from ..dataframe_utils import append_df, merge_df, resample_whole_df

LOG = logging.getLogger("metloom.pointdata.cdec")


class CDECPointData(PointData):
    """
    Implement PointData methods for CDEC data source
    API documentation here https://cdec.water.ca.gov/dynamicapp/
    """

    ALLOWED_VARIABLES = CdecStationVariables
    CDEC_URL = "http://cdec.water.ca.gov/dynamicapp/req/JSONDataServlet"
    META_URL = "https://cdec.water.ca.gov/dynamicapp/req/CSVMetaDataServlet"
    DATASOURCE = "CDEC"

    def __init__(self, station_id, name, metadata=None):
        """
        See docstring for PointData.__init__
        """
        super(CDECPointData, self).__init__(station_id, name, metadata=metadata)
        self._raw_metadata = None
        # CDEC has datetimes that aren't found in US/Pacific. Use this instead
        self._tzinfo = timezone(timedelta(hours=-8.0))

    def _get_all_metadata(self):
        """
        Get all the raw metadata for a station. This is a list of sensor
        descriptions for the station
        Returns:
            A list of dictionaries describing the sensors at a station
        """
        if self._raw_metadata is None:
            url = self.META_URL + f"?Stations={self.id}"
            resp = requests.get(url)
            resp.raise_for_status()
            result = StringIO(resp.text)
            self._raw_metadata = pd.read_csv(result)
        return self._raw_metadata

    def _check_snowcourse_sensors(self):
        """
        Returns a list of booleans for snow sensors that checks if they
        are monthly in cadence
        """
        snow_sensors = [18, 3, 82]
        data = self._get_all_metadata()
        manual_check = []
        for num, d in zip(data["SENSOR_NUMBER"], data["DURATION"]):
            if int(num) in snow_sensors:
                manual_check.append(d == "M")
        return manual_check

    def is_only_snow_course(self, variables: List[SensorDescription]):
        """
        Determine if a station only has snow course measurements
        """
        manual_check = self._check_snowcourse_sensors()

        result = False
        if len(manual_check) > 0 and all(manual_check):
            result = True
        if result and not self.is_only_monthly():
            # This would happen if all snow sensors have code "M"
            # but there are other hourly or daily sensors
            if any(
                [sensor not in [
                    self.ALLOWED_VARIABLES.SWE, self.ALLOWED_VARIABLES.SNOWDEPTH
                ] for sensor in variables]
            ):
                # this is an acceptable scenario where we are looking for
                # more than just snow variables and the station has them
                result = False
            else:
                # We are only looking for snow variables and this only has
                # monthly snow variables so it is effectively only
                # a snow course
                result = True
        return result

    def is_partly_snow_course(self):
        """
        Determine if any of the snow sensors at a station are on a monthly
        interval
        Assumption: Monthly snow sensor measurements are snow courses
        """
        result = self._check_snowcourse_sensors()
        return any(result)

    def is_only_monthly(self):
        """
        determine if all sensors for a station are on a monthly interval
        """
        data = self._get_all_metadata()
        duration = data["DURATION"].values
        manual_check = [d == "M" for d in duration]
        if len(manual_check) > 0 and all(manual_check):
            return True
        return False

    def _get_metadata(self):
        """
        See docstring for PointData._get_metadata
        """
        data = self._get_all_metadata()
        chosen_sensor_data = data.iloc[0]
        sensor_nums = data["SENSOR_NUMBER"].values

        # go down a rank of choice sensor metadata
        for choice in [3, 18, 2, 30]:
            if choice in sensor_nums:
                chosen_sensor_data = data.loc[
                    data["SENSOR_NUMBER"] == choice
                ].iloc[0]
                break

        return gpd.points_from_xy(
            [float(chosen_sensor_data["LONGITUDE"])],
            [float(chosen_sensor_data["LATITUDE"])],
            z=[float(chosen_sensor_data["ELEVATION"])],
        )[0]

    def _data_request(self, params):
        """
        Make get request to CDEC and return JSON
        Args:
            params: dictionary of request parameters
        Returns:
            dictionary of response values
        """
        resp = requests.get(self.CDEC_URL, params=params)
        resp.raise_for_status()
        return resp.json()

    def _sensor_response_to_df(self, response_data, sensor, final_columns,
                               resample_duration=None):
        """
        Convert the response data from the API to a GeoDataFrame
        Format and map columns in the dataframe
        Args:
            response_data: JSON list response from CDEC API
            sensor: SensorDescription obj
            final_columns: List of columns used for filtering
            resample_duration: duration to resample to
        Returns:
            GeoDataFrame
        """
        sensor_df = gpd.GeoDataFrame.from_dict(
            response_data,
            geometry=[self.metadata] * len(response_data),
        )
        sensor_df.replace(-9999.0, np.nan, inplace=True)
        # this mapping is important. Sometimes obsDate is null
        column_map = {
            "date": "datetime",
            "value": sensor.name,
            "units": f"{sensor.name}_units",
            "stationId": "site",
        }
        if "measurementDate" in final_columns:
            column_map["obsDate"] = "measurementDate"

        sensor_df.rename(
            columns=column_map,
            inplace=True,
        )
        final_columns += [sensor.name, f"{sensor.name}_units"]
        sensor_df["datetime"] = pd.to_datetime(sensor_df["datetime"])

        # resample if necessary
        if resample_duration:
            sensor_df = resample_whole_df(
                sensor_df.set_index("datetime"), sensor,
                interval=resample_duration
            ).reset_index()
            sensor_df = GeoDataFrame(sensor_df, geometry=sensor_df["geometry"])

        sensor_df["datetime"] = sensor_df["datetime"].apply(self._handle_df_tz)
        if "measurementDate" in sensor_df.columns:
            sensor_df["measurementDate"] = pd.to_datetime(sensor_df["measurementDate"])
            sensor_df["measurementDate"] = sensor_df["measurementDate"].apply(
                self._handle_df_tz
            )
        # set index so joining works
        sensor_df.set_index("datetime", inplace=True)
        sensor_df = sensor_df.filter(final_columns)
        sensor_df = sensor_df.loc[pd.notna(sensor_df[sensor.name])]
        return sensor_df

    def _get_data_fallback(self, params, duration_list):
        """
        Allow for fallback on finer resolution API durations with resample
        if the desired duration does not return data
        Args:
            params: request params with or without dur_code
            duration_list: list of durations to try. First index is desired
                durations
        """
        if len(duration_list) < 1:
            raise ValueError("Duration list cannot be empty")
        response_data = []
        df_duration = duration_list[0]
        for duration in duration_list:
            params["dur_code"] = duration
            response_data = self._data_request(params)
            if response_data:
                df_duration = duration
                break
        return response_data, df_duration

    def _get_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
        duration_list: List[str],
        include_measurement_date=False
    ):
        """
        Args:
            start_date: datetime object for start of data collection period
            end_date: datetime object for end of data collection period
            variables: List of metloom.variables.SensorDescription object
                from self.ALLOWED_VARIABLES
            duration_list: CDEC duration code and fallbacks ['D', 'H', 'E']
            include_measurement_date: boolean for including the
                'measurmentDate' column in the resulting dataframe. This column
                is only relevant for snow courses
        Returns:
            GeoDataFrame of data, indexed on datetime, site
        """
        params = {
            "Stations": self.id,
            "Start": start_date.isoformat(),
            "End": end_date.isoformat(),
        }
        df = None
        final_columns = ["geometry", "site"]
        desired_duration = duration_list[0]
        if include_measurement_date:
            final_columns += ["measurementDate"]
        for sensor in variables:
            params["SensorNums"] = sensor.code
            response_data, response_duration = self._get_data_fallback(
                params, duration_list
            )
            if response_data:
                # don't resample if we have the desired duration
                if response_duration == desired_duration:
                    resample_duration = None
                else:
                    resample_duration = desired_duration
                sensor_df = self._sensor_response_to_df(
                    response_data, sensor, final_columns,
                    resample_duration=resample_duration
                )
                df = merge_df(df, sensor_df)

        if df is not None:
            if len(df.index) > 0:
                # Set the datasource
                df["datasource"] = [self.DATASOURCE] * len(df.index)
                df.reset_index(inplace=True)
                df.set_index(keys=["datetime", "site"], inplace=True)
                df.index.set_names(["datetime", "site"], inplace=True)
            else:
                df = None
        self.validate_sensor_df(df)
        return df

    def get_event_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        return self._get_data(start_date, end_date, variables, ["E"])

    def get_daily_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_daily_data
        Example query:
        https://cdec.water.ca.gov/dynamicapp/req/JSONDataServlet?
        Stations=TNY&SensorNums=3&dur_code=D&Start=2021-05-16&End=2021-05-16
        """
        return self._get_data(start_date, end_date, variables, ["D", "h", "E"])

    def get_hourly_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_hourly_data
        """
        return self._get_data(start_date, end_date, variables, ["h", "E"])

    def get_snow_course_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_snow_course_data
        """
        if not self.is_partly_snow_course():
            raise ValueError(f"{self.id} is not a snow course")
        return self._get_data(
            start_date, end_date, variables, ["M"], include_measurement_date=True
        )

    @staticmethod
    def _station_sensor_search(
        bounds, sensor: SensorDescription, dur=None, collect=None,
        buffer=0.0
    ):
        """
        Station search form https://cdec.water.ca.gov/dynamicapp/staSearch?
        Search for stations using the CDEC station search utility
        Args:
            bounds: dictionary of Longitude and Latitidue bounds with keys
                minx, maxx, miny, maxy
            sensor: SensorDescription object
            dur: optional CDEC duration code ['M', 'H', 'D']
            collect: optional CDEC collection type string i.e. 'MANUAL+ENTRY'
        Returns:
            Pandas Dataframe of table result or None if no table found

        """
        dur_str = f"&dur_chk=on&dur={dur}" if dur else "&dur="
        collect_str = (
            f"&collect_chk=on&collect={collect}"
            if collect
            else "&collect=NONE+SPECIFIED"
        )
        url = (
            f"https://cdec.water.ca.gov/dynamicapp/staSearch?sta="
            f"&sensor_chk=on&sensor={sensor.code}"
            f"{collect_str}"
            f"{dur_str}"
            f"&active_chk=on&active=Y"
            f"&loc_chk=on"
            f"&lon1={bounds['minx']-buffer}&lon2={bounds['maxx']+buffer}"
            f"&lat1={bounds['miny']-buffer}&lat2={bounds['maxy']+buffer}"
            f"&elev1=-5&elev2=99000&nearby=&basin=NONE+SPECIFIED"
            f"&hydro=NONE+SPECIFIED&county=NONE+SPECIFIED&agency_num=160"
            f"&display=sta"
        )
        try:
            return pd.read_html(url)[0]
        except ValueError:
            LOG.error(f"No tables for {url}")
            return None

    @classmethod
    def points_from_geometry(
        cls,
        geometry: gpd.GeoDataFrame,
        variables: List[SensorDescription],
        **kwargs
    ):
        """
        See docstring for PointData.points_from_geometry

        Args:
            geometry: GeoDataFrame for shapefile from gpd.read_file
            variables: List of SensorDescription
            snow_courses: Boolean for including only snowcourse data or no
            snowcourse data
            within_geometry: filter the points to within the shapefile
            instead of just the extents. Default True
            buffer: buffer added to search box

        Returns:
            PointDataCollection
        """
        # assign defaults
        kwargs = cls._add_default_kwargs(kwargs)

        # Assume station search result is in 4326
        projected_geom = geometry.to_crs(4326)
        bounds = projected_geom.bounds.iloc[0]
        search_df = None
        station_search_kwargs = {}

        # Filter to manual, monthly measurements if looking for snow courses
        if kwargs['snow_courses']:
            station_search_kwargs["dur"] = "M"
            station_search_kwargs["collect"] = "MANUAL+ENTRY"
        for variable in variables:
            result_df = cls._station_sensor_search(
                bounds, variable, buffer=kwargs["buffer"],
                **station_search_kwargs
            )
            if result_df is not None:
                result_df["index_id"] = result_df["ID"]
                result_df.set_index("index_id", inplace=True)
                search_df = append_df(
                    search_df, result_df
                ).drop_duplicates(subset=['ID'])
        # return empty collection if we didn't find any points
        if search_df is None:
            return cls.ITERATOR_CLASS([])
        clms = search_df.columns.values
        if "ElevationFeet" in clms:
            elev_key = "ElevationFeet"
        elif "Elevation Feet" in clms:
            elev_key = "Elevation Feet"
        else:
            raise RuntimeError("No key for elevation")
        gdf = gpd.GeoDataFrame(
            search_df,
            geometry=gpd.points_from_xy(
                search_df["Longitude"],
                search_df["Latitude"],
                z=search_df[elev_key],
            ),
        )
        # filter to points within shapefile
        if kwargs['within_geometry']:
            filtered_gdf = gdf[gdf.within(projected_geom.iloc[0]["geometry"])]
        else:
            filtered_gdf = gdf

        points = [
            cls(row[0], row[1], metadata=row[2])
            for row in zip(
                filtered_gdf.index,
                filtered_gdf["Station Name"],
                filtered_gdf["geometry"],
            )
        ]
        # filter to snow courses or not snowcourses depending on desired result
        if kwargs['snow_courses']:
            return cls.ITERATOR_CLASS([p for p in points if p.is_partly_snow_course()])
        else:
            return cls.ITERATOR_CLASS(
                [p for p in points if not p.is_only_snow_course(variables)]
            )



from datetime import datetime
import pandas as pd
import zeep

from metloom.request_utils import no_ssl_verification


class BaseSnotelClient:
    """
    Base snotel client class. Used for interacting with SNOTEL SOAP client.
    This is just a base class and not meant for direct use.

    Example use with extended class::

        MetaDataSnotelClient('TNY:CA:SNOW').get_data()

    """
    URL = "https://wcc.sc.egov.usda.gov/awdbWebService/services?WSDL"
    # map allowed params to API filter params
    PARAMS_MAP = {
        'station_triplet': 'stationTriplet',
        'station_triplets': 'stationTriplets',
        'begin_date': 'beginDate',
        'end_date': 'endDate',
        'max_longitude': 'maxLongitude',
        'min_longitude': 'minLongitude',
        'max_latitude': 'maxLatitude',
        'min_latitude': 'minLatitude',
        'network_cds': 'networkCds',
        "duration": "duration",
        'ordinal': 'ordinal',
        'element_cd': 'elementCd',
        'element_cds': 'elementCds',
        'parameter': 'parameter',
        'height_depth': 'heightDepth',
    }
    SERVICE_NAME = None
    DEFAULT_PARAMS = {}

    def __init__(self, **kwargs):
        self.params = self._get_params(**kwargs)
        if getattr(self, "DURATION", None) is not None:
            extra_params = {"duration": self.DURATION, **self.DEFAULT_PARAMS}
        else:
            extra_params = self.DEFAULT_PARAMS
        # add in default params to self.params, don't replace user entries
        for key, value in extra_params.items():
            if key not in self.params:
                self.params[key] = value

    def _get_params(self, **kwargs):
        """
        map input parameter keys to keys expected by the SOAP client
        """
        params = {}
        for key, value in kwargs.items():
            mapped_key = self.PARAMS_MAP.get(key)
            if mapped_key is None:
                raise ValueError(
                    f"Could not find valid mapped key for {key}"
                )
            if isinstance(value, datetime):
                value = value.date().isoformat()
            params.update({mapped_key: value})
        return params

    @classmethod
    def _make_request(cls, **params):
        """
        Make the request to the SOAP client for the implemented service.
        """
        with no_ssl_verification():
            client = zeep.Client(
                cls.URL,
            )
            service = getattr(client.service, cls.SERVICE_NAME)
            response = service(**params)
        return response

    def get_data(self):
        """
        Make the actual request and return data.
        """
        data = self._make_request(**self.params)
        return data


class MetaDataSnotelClient(BaseSnotelClient):
    """
    Read metadata from the metadata service for a particular station triplet
    """
    SERVICE_NAME = "getStationMetadata"

    def __init__(self, station_triplet: str, **kwargs):
        super(MetaDataSnotelClient, self).__init__(
            station_triplet=station_triplet, **kwargs
        )

    def get_data(self):
        """
        Returns a dictionary of metadata values
        """
        data = self._make_request(**self.params)
        # change ordered dict of values to regular dict
        return dict(data.__values__)


class ElementSnotelClient(BaseSnotelClient):
    """
    Get all station elements for a station triplet. Station triplets
    are descriptions of each sensor on the station

    get_data returns a list of zeep objects. Zeep objects are indexible
    or attributes can be accessed with getattr or ``.``
    """
    SERVICE_NAME = "getStationElements"

    def __init__(self, station_triplet: str, **kwargs):
        super(ElementSnotelClient, self).__init__(
            station_triplet=station_triplet, **kwargs
        )


class PointSearchSnotelClient(BaseSnotelClient):
    """
    Search for stations based on criteria. This search is default logical
    ``AND`` meaning all criteria need to be true.
    get_data returns a list of string station triplets
    """
    SERVICE_NAME = 'getStations'
    DEFAULT_PARAMS = {
        'logicalAnd': 'true',
    }

    def __init__(self, max_latitude: float, min_latitude: float,
                 max_longitude: float, min_longitude: float,
                 network_cds: str, element_cds: str, **kwargs):
        super(PointSearchSnotelClient, self).__init__(
            max_latitude=max_latitude, min_latitude=min_latitude,
            max_longitude=max_longitude, min_longitude=min_longitude,
            network_cds=network_cds, element_cds=element_cds,
            **kwargs
        )


class SeriesSnotelClient(BaseSnotelClient):
    """
    Base extension for services that return timseries data.
    """
    SERVICE_NAME = "getData"
    DURATION = "DAILY"
    DEFAULT_PARAMS = {
        "ordinal": 1,
        "getFlags": "true",
        "alwaysReturnDailyFeb29": "false",
    }

    def __init__(self, begin_date: datetime, end_date: datetime,
                 station_triplet: str, **kwargs):
        super(SeriesSnotelClient, self).__init__(
            begin_date=begin_date, end_date=end_date,
            station_triplets=[station_triplet], **kwargs)

    @staticmethod
    def _parse_data(raw_data):
        """
        Parse the return data to return a consistent format for timeseries
        data
        """
        data = raw_data[0]
        mapped_data = []
        collection_dates = getattr(data, "collectionDates", None)
        if len(data["values"]) == 0:
            return mapped_data
        if collection_dates:
            date_list = [pd.to_datetime(d) for d in collection_dates]
        else:
            date_list = pd.date_range(data["beginDate"], data["endDate"])

        for date_obj, flag, value in zip(date_list, data["flags"], data["values"]):
            if date_obj is not None:
                mapped_data.append({
                    "datetime": date_obj,
                    "flag": flag,
                    "value": float(value) if value is not None else None
                })
        return mapped_data

    def get_data(self, element_cd: str, **extra_params):
        """
        get the timeseires data

        Args:
            element_cd: the variable code from the allowed variable codes
                in the API
            extra_params: kwargs for any extra parameters. These override
                the default parameters
        """
        extra_params.update(element_cd=element_cd)
        mapped_params = self._get_params(**extra_params)
        params = {**self.params, **mapped_params}
        data = self._make_request(**params)
        return self._parse_data(data)


class DailySnotelDataClient(SeriesSnotelClient):
    """
    Class for getting daily data
    """
    pass


class SemiMonthlySnotelClient(SeriesSnotelClient):
    """
    Class for getting semi monthly (snow course) data
    """
    DURATION = "SEMIMONTHLY"


class HourlySnotelDataClient(SeriesSnotelClient):
    """
    Class for getting hourly data
    """
    DURATION = None
    SERVICE_NAME = "getHourlyData"
    DEFAULT_PARAMS = {
        "ordinal": 1,
    }

    @staticmethod
    def _parse_data(raw_data):
        """
        Clean the hourly data to be consistent with timeseries results
        """
        data = raw_data[0]
        mapped_data = []
        for row in data["values"]:
            value = row["value"]
            mapped_data.append({
                "datetime": pd.to_datetime(row["dateTime"]),
                "flag": row["flag"],
                "value": float(value) if value is not None else None
            })
        return mapped_data



from datetime import datetime, timezone, timedelta
from typing import List, Dict
import logging
import geopandas as gpd
import pandas as pd
from functools import reduce

from .base import PointData
from ..variables import SnotelVariables, SensorDescription
from ..dataframe_utils import append_df, merge_df

from .snotel_client import (
    DailySnotelDataClient, MetaDataSnotelClient, HourlySnotelDataClient,
    SemiMonthlySnotelClient, PointSearchSnotelClient, SeriesSnotelClient,
    ElementSnotelClient
)

LOG = logging.getLogger("metloom.pointdata.snotel")


class SnotelPointData(PointData):
    """
    Implement PointData methods for SNOTEL data source
    API documentation here:
    https://www.nrcs.usda.gov/wps/portal/wcc/home/dataAccessHelp/webService/webServiceReference/
    Website has variable and network codes
    Possible testing resource:
    https://wcc.sc.egov.usda.gov/awdbWebService/webservice/testwebservice.jsf?webserviceName=/awdbWebService
    """

    ALLOWED_VARIABLES = SnotelVariables
    DATASOURCE = "NRCS"

    def __init__(self, station_id, name, metadata=None):
        """
        See docstring for PointData.__init__
        """
        super(SnotelPointData, self).__init__(
            station_id, name, metadata=metadata
        )
        self._raw_metadata = None
        self._raw_elements = None
        self._tzinfo = None

    def _snotel_response_to_df(self, result_map: Dict[SensorDescription, List[dict]],
                               duration: str, include_measurement_date=False):
        """
        Convert the response from climata.snotel classes into
        Args:
        result_map: map of the sensors to the list of API results
        duration: string representation of the duration tag for the
            API (i.e. HOURLY)
        include_measurement_date: boolean for including the
            'measurementDate' column in the resulting dataframe. This column
            is only relevant for snow courses

        """
        df = None
        # TODO: possible DRY opportunity here too
        final_columns = ["geometry", "site"]
        if include_measurement_date:
            final_columns += ["measurementDate"]

        for variable, data in result_map.items():
            transformed = []
            for row in data:
                row_obj = {
                    "datetime": row["datetime"],
                    "site": self.id,
                    variable.name: row["value"],
                    f"{variable.name}_units": self._get_units(variable, duration),
                }
                if include_measurement_date:
                    row_obj["measurementDate"] = row["datetime"]
                transformed.append(row_obj)

            final_columns += [variable.name, f"{variable.name}_units"]
            sensor_df = gpd.GeoDataFrame.from_dict(
                transformed, geometry=[self.metadata] * len(transformed)
            )

            sensor_df = sensor_df.drop_duplicates()

            # TODO: possibly an opportunity for DRY here (see CDEC)
            sensor_df["datetime"] = pd.to_datetime(sensor_df["datetime"])
            sensor_df["datetime"] = sensor_df["datetime"].apply(
                self._handle_df_tz
            )
            if include_measurement_date:
                sensor_df["measurementDate"] = pd.to_datetime(
                    sensor_df["measurementDate"]
                )
                sensor_df["measurementDate"] = sensor_df["measurementDate"].apply(
                    self._handle_df_tz
                )
            # set index so joining works
            sensor_df.set_index("datetime", inplace=True)
            sensor_df = sensor_df.filter(final_columns)
            # filter to rows that have value
            sensor_df = sensor_df.loc[pd.notna(sensor_df[variable.name])]
            df = merge_df(df, sensor_df)

        if df is not None and len(df.index) > 0:
            df["datasource"] = [self.DATASOURCE] * len(df.index)
            df.reset_index(inplace=True)
            df.set_index(keys=["datetime", "site"], inplace=True)
            df.index.set_names(["datetime", "site"], inplace=True)
        self.validate_sensor_df(df)
        return df

    def _fetch_data_for_variables(self, client: SeriesSnotelClient,
                                  variables: List[SensorDescription],
                                  duration: str,
                                  include_measurement_date=False,
                                  ):
        result_map = {}
        for variable in variables:
            params = variable.extra or {}
            data = client.get_data(element_cd=variable.code, **params)
            if len(data) > 0:
                result_map[variable] = data
            else:
                LOG.warning(f"No {variable.name} found for {self.name}")
        return self._snotel_response_to_df(
            result_map, duration, include_measurement_date=include_measurement_date
        )

    def get_daily_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_daily_data
        """
        client = DailySnotelDataClient(
            station_triplet=self.id,
            begin_date=start_date,
            end_date=end_date,
        )
        return self._fetch_data_for_variables(client, variables,
                                              client.DURATION)

    def get_hourly_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription]
    ):
        """
        See docstring for PointData.get_hourly_data
        """

        client = HourlySnotelDataClient(
            station_triplet=self.id,
            begin_date=start_date,
            end_date=end_date,
        )
        return self._fetch_data_for_variables(client, variables, "HOURLY")

    def get_snow_course_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_snow_course_data
        """
        client = SemiMonthlySnotelClient(
            station_triplet=self.id,
            begin_date=start_date,
            end_date=end_date,
        )
        return self._fetch_data_for_variables(
            client, variables, client.DURATION, include_measurement_date=True
        )

    def _get_all_metadata(self):
        """
        Set _raw_metadata once using Snotel API
        """
        if self._raw_metadata is None:
            client = MetaDataSnotelClient(station_triplet=self.id)
            self._raw_metadata = client.get_data()
        return self._raw_metadata

    def _get_all_elements(self):
        """
        Set _raw_metadata once using Snotel API
        """
        if self._raw_elements is None:
            client = ElementSnotelClient(station_triplet=self.id)
            self._raw_elements = client.get_data()
        return self._raw_elements

    def _get_units(self, variable: SensorDescription, duration: str):
        units = None
        for meta in self._get_all_elements():
            if meta["elementCd"] == variable.code and meta["duration"] == duration:
                units = meta["storedUnitCd"]
                break
        if units is None:
            raise ValueError(f"Could not find units for {variable}")
        return units

    def _get_metadata(self):
        """
        See docstring for PointData._get_metadata
        """
        all_metadata = self._get_all_metadata()
        if isinstance(all_metadata, list):
            data = all_metadata[0]
        else:
            data = all_metadata
        return gpd.points_from_xy(
            [data["longitude"]], [data["latitude"]], z=[data["elevation"]]
        )[0]

    def _get_tzinfo(self):
        """
        Return timezone info that pandas can use from the raw_metadata
        """
        metadata = self._get_all_metadata()
        # Snow courses might not have a timezone attached
        tz_hours = metadata.get("stationDataTimeZone")
        if tz_hours is None:
            LOG.error(f"Could not find timezone info for {self.id} ({self.name})")
            tz_hours = 0
        else:
            tz_hours = float(tz_hours)
        return timezone(timedelta(hours=tz_hours))

    @property
    def tzinfo(self):
        if self._tzinfo is None:
            self._tzinfo = self._get_tzinfo()
        return self._tzinfo

    @classmethod
    def points_from_geometry(
        cls,
        geometry: gpd.GeoDataFrame,
        variables: List[SensorDescription],
        **kwargs
    ):
        """
        See docstring for PointData.points_from_geometry

        Args:
            geometry: GeoDataFrame for shapefile from gpd.read_file
            variables: List of SensorDescription
            snow_courses: boolean for including only snowcourse data or no
                snowcourse data
            within_geometry: filter the points to within the shapefile
                instead of just the extents. Default True
            buffer: buffer added to search box
        Returns:
            PointDataCollection
        """
        # assign defaults
        kwargs = cls._add_default_kwargs(kwargs)

        projected_geom = geometry.to_crs(4326)
        bounds = projected_geom.bounds.iloc[0]
        # TODO: network may need to change to get streamflow
        network = "SNOW" if kwargs['snow_courses'] else [
            "SNTL", "USGS", "BOR", "COOP", "SNTLT"
        ]
        point_codes = []
        buffer = kwargs["buffer"]
        search_kwargs = {
            "max_latitude": bounds["maxy"] + buffer,
            "min_latitude": bounds["miny"] - buffer,
            "max_longitude": bounds["maxx"] + buffer,
            "min_longitude": bounds["minx"] - buffer,
            "network_cds": network,
        }
        for variable in variables:
            # this search is default AND on all parameters
            # so search for each variable seperately
            response = PointSearchSnotelClient(
                element_cds=variable.code,
                **search_kwargs
                # ordinals=  # TODO: what are ordinals?
                # ordinals are NRCS descriptors for parameters that may have multiple
                # measurements, such as two pillows at a site, or two measurements of
                # SWE from the same pillow. Fairly comfortable with the idea that we
                # can assume ordinal=1
            ).get_data()
            if len(response) > 0:
                point_codes += response

        # no duplicate codes
        point_codes = list(set(point_codes))
        dfs = [
            pd.DataFrame.from_records(
                [MetaDataSnotelClient(station_triplet=code).get_data()]
            ).set_index("stationTriplet") for code in point_codes
        ]

        if len(dfs) > 0:
            df = reduce(lambda a, b: append_df(a, b), dfs)
        else:
            return cls.ITERATOR_CLASS([])

        df.reset_index(inplace=True)
        gdf = gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(
                df["longitude"], df["latitude"], z=df["elevation"]
            ),
        )
        if kwargs['within_geometry']:
            filtered_gdf = gdf[gdf.within(projected_geom.iloc[0]["geometry"])]
        else:
            filtered_gdf = gdf
        points = [
            cls(row[0], row[1], metadata=row[2])
            for row in zip(
                filtered_gdf["stationTriplet"],
                filtered_gdf["name"],
                filtered_gdf["geometry"],
            )
        ]
        return cls.ITERATOR_CLASS(points)



import copy
import logging
from datetime import datetime
from typing import List

import pandas as pd

import geopandas as gpd

from ..variables import SensorDescription, VariableBase

LOG = logging.getLogger("metloom.pointdata.base")


class DataValidationError(RuntimeError):
    pass


class PointDataCollection:
    """
    Iterator class for a collection of PointData objects.
    This allows conversion to a GeoDataFrame
    """

    def __init__(self, points: List[object] = None):
        """

        Args:
            points: List of point data objects
        """
        self.points = points or []
        self._index = 0

    def add_point(self, point):
        """
        Append point to collection of PointData objects

        Args:
            point: PointData object
        """
        self.points.append(point)

    def to_dataframe(self):
        """
        Returns:
            GeoDataFrame of points. Columns are ['name', 'id', 'geometry']
        """
        names = []
        ids = []
        meta = []
        datasource = []
        for point in self.points:
            names += [point.name]
            ids += [point.id]
            meta += [point.metadata]
            datasource += [point.DATASOURCE]

        obj = {"name": names, "id": ids, "datasource": datasource}
        return gpd.GeoDataFrame.from_dict(obj, geometry=meta)

    def __len__(self):
        return len(self.points)

    def __iter__(self):
        for item in self.points:
            yield item


class GenericPoint(object):
    """
    Class for storing metadata. and defining the expected data format
    returned from `get_data` methods

    """
    ALLOWED_VARIABLES = VariableBase
    ITERATOR_CLASS = PointDataCollection
    DATASOURCE = None
    EXPECTED_COLUMNS = ["geometry", "datasource"]
    EXPECTED_INDICES = ["datetime", "site"]
    NON_VARIABLE_COLUMNS = EXPECTED_INDICES + EXPECTED_COLUMNS

    def __init__(self, station_id, name, metadata=None):
        """

        Args:
            station_id: code used within datasource API to access station
            name: station name. This will be used in the GeoDataFrames
            metadata: Optional shapely point. This will bypass the
                _get_metadata method if provided
        """
        self.id = station_id
        self.name = name
        self._metadata = metadata
        self.desired_tzinfo = "UTC"

    def _get_metadata(self):
        """
        Method to get a shapely Point object to describe the station location

        Returns:
            shapely.point.Point object in Longitude, Latitude
        """
        raise NotImplementedError("_get_metadata is not implemented")

    def _handle_df_tz(self, val):
        """
        Covert one entry from a df from cls.TZINFO to UTC
        """
        if pd.isna(val):
            return val
        else:
            local = val.tz_localize(self.tzinfo)
            return local.tz_convert(self.desired_tzinfo)

    @property
    def tzinfo(self):
        """
        tzinfo that pandas can use for tz_localize
        """
        return self._tzinfo

    @property
    def metadata(self):
        """
        metadata property
        Returns:
            shapely.point.Point object in Longitude, Latitude with z in ft
        """
        if self._metadata is None:
            self._metadata = self._get_metadata()
        return self._metadata

    @classmethod
    def validate_sensor_df(cls, gdf: gpd.GeoDataFrame):
        """
        Validate that the GeoDataFrame returned is formatted correctly.
        The goal of this method is to ensure base classes are returning a
        consistent format of dataframe
        """
        if gdf is None:
            return
        assert isinstance(gdf, gpd.GeoDataFrame)
        columns = gdf.columns
        index_names = gdf.index.names
        # check for required indexes
        for ei in cls.EXPECTED_INDICES:
            if ei not in index_names:
                raise DataValidationError(
                    f"{ei} was expected, but not found as an"
                    f" index of the final dataframe"
                )
        # check for expected columns - avoid modifying at class level
        expected_columns = copy.deepcopy(cls.EXPECTED_COLUMNS)
        possible_extras = ["measurementDate", "quality_code"]
        for pe in possible_extras:
            if pe in columns:
                expected_columns += [pe]
        for column in expected_columns:
            if column not in columns:
                raise DataValidationError(
                    f"{column} was expected, but not found as a"
                    f" column of the final dataframe"
                )

        remaining_columns = [c for c in columns if c not in expected_columns]
        # make sure all variables have a units column as well
        for rc in remaining_columns:
            if "_units" not in rc:
                assert f"{rc}_units" in remaining_columns

    def __repr__(self):
        return f"{self.__class__.__name__}({self.id!r}, {self.name!r})"

    def __str__(self):
        return f"{self.name} ({self.id})"

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.id == other.id and self.name == other.name


class PointData(GenericPoint):
    """
    Extend GenericPoint and add functions for finding data from geometry
    and for gettings daily, hourly, or snow course data
    """

    # Default kwargs for function points from geometry
    POINTS_FROM_GEOM_DEFAULTS = {
        'within_geometry': True, 'snow_courses': False,
        'buffer': 0.0, "filter_to_active": False
    }

    def get_daily_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        Get daily measurement data
        Args:
            start_date: datetime object for start of data collection period
            end_date: datetime object for end of data collection period
            variables: List of metloom.variables.SensorDescription object
                from self.ALLOWED_VARIABLES
        Returns:
            GeoDataFrame of data. The dataframe should be indexed on
            ['datetime', 'site'] and have columns
            ['geometry', 'site', 'measurementDate']. Additionally, for each
            variables, it should have column f'{variable.name}' and
            f'{variable.name}_UNITS'
            See CDECPointData._get_data for example implementation and
            TestCDECStation.tny_daily_expected for example dataframe.
            Datetimes should be in UTC
        """
        raise NotImplementedError("get_daily_data is not implemented")

    def get_hourly_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        Get hourly measurement data
        Args:
            start_date: datetime object for start of data collection period
            end_date: datetime object for end of data collection period
            variables: List of metloom.variables.SensorDescription object
                from self.ALLOWED_VARIABLES
        Returns:
            GeoDataFrame of data. The dataframe should be indexed on
            ['datetime', 'site'] and have columns
            ['geometry', 'site', 'measurementDate']. Additionally, for each
            variables, it should have column f'{variable.name}' and
            f'{variable.name}_UNITS'
            See CDECPointData._get_data for example implementation and
            TestCDECStation.tny_daily_expected for example dataframe.
            Datetimes should be in UTC
        """
        raise NotImplementedError("get_hourly_data is not implemented")

    def get_snow_course_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        Get snow course data
        Args:
            start_date: datetime object for start of data collection period
            end_date: datetime object for end of data collection period
            variables: List of metloom.variables.SensorDescription object
                from self.ALLOWED_VARIABLES
        Returns:
            GeoDataFrame of data. The dataframe should be indexed on
            ['datetime', 'site'] and have columns
            ['geometry', 'site', 'measurementDate']. Additionally, for each
            variables, it should have column f'{variable.name}' and
            f'{variable.name}_UNITS'
            See CDECPointData._get_data for example implementation and
            TestCDECStation.tny_daily_expected for example dataframe.
            Datetimes should be in UTC
        """
        raise NotImplementedError("get_snow_course_data is not implemented")

    @classmethod
    def _add_default_kwargs(cls, kwargs):
        """
        Populates the kwargs for the points from geometry function
        """
        for k, v in cls.POINTS_FROM_GEOM_DEFAULTS.items():
            if k not in kwargs.keys():
                kwargs[k] = v
        return kwargs

    @classmethod
    def points_from_geometry(
        cls,
        geometry: gpd.GeoDataFrame,
        variables: List[SensorDescription],
        snow_courses=False,
        within_geometry=True,
        buffer=0.0
    ):
        """
        Find a collection of points with measurements for certain variables
        contained within a shapefile. Any point in the shapefile with
        measurements for any of the variables should be included
        Args:
            geometry: GeoDataFrame for shapefile from gpd.read_file
            variables: List of SensorDescription
            snow_courses: boolean for including only snowcourse data or no
                snowcourse data
            within_geometry: filter the points to within the shapefile
                instead of just the extents. Default True
            buffer: buffer added to search box
        Returns:
            PointDataCollection
        """
        raise NotImplementedError("points_from_geometry not implemented")

    @classmethod
    def _validate_geodataframe(cls, gdf: gpd.GeoDataFrame):
        if not isinstance(gdf, gpd.GeoDataFrame):
            raise DataValidationError('Returned DataFrame must be a GeoDataframe')

    @classmethod
    def _validate_df_indicies(cls, gdf: gpd.GeoDataFrame):
        """ Confirm the df is indexed properly"""
        for ei in cls.EXPECTED_INDICES:
            if ei not in gdf.index.names:
                raise DataValidationError(
                    f"{ei} was expected, but not found as an"
                    f" index of the final dataframe"
                )

    @classmethod
    def _validate_df_columns(cls, gdf: gpd.GeoDataFrame, expected_columns: List[str]):
        # check for expected columns - avoid modifying at class level
        possible_extras = ["measurementDate", "quality_code"]
        columns = gdf.columns
        for pe in possible_extras:
            if pe in columns:
                expected_columns += [pe]

        for column in expected_columns:
            if column not in columns:
                raise DataValidationError(
                    f"{column} was expected, but not found as a"
                    f" column of the final dataframe"
                )

    @classmethod
    def _validate_df_units(cls, gdf: gpd.GeoDataFrame, expected_columns: List[str]):
        """
        Check the variables requested have units associated
        """
        remaining_columns = [c for c in gdf.columns if c not in expected_columns]
        # make sure all variables have a units column as well
        for rc in remaining_columns:
            if "_units" not in rc:
                if f"{rc}_units" not in remaining_columns:
                    raise DataValidationError(f'Missing units column for {rc}')

    @classmethod
    def validate_sensor_df(cls, gdf: gpd.GeoDataFrame):
        """
        Validate that the GeoDataFrame returned is formatted correctly.
        The goal of this method is to ensure base classes are returning a
        consistent format of dataframe
        """
        if gdf is None:
            return

        expected_columns = copy.deepcopy(cls.EXPECTED_COLUMNS)

        # Confirm the dataframe is a geodataframe
        cls._validate_geodataframe(gdf)
        # Confirm the df is indexed properly
        cls._validate_df_indicies(gdf)
        # Confirm the columns are correct
        cls._validate_df_columns(gdf, expected_columns)
        # Confirm that any columns from variables have associated units
        cls._validate_df_units(gdf, expected_columns)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.id!r}, {self.name!r})"

    def __str__(self):
        return f"{self.name} ({self.id})"

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.id == other.id and self.name == other.name
